{$A+,B-,C+,E-,F-,G+,H+,I+,J-,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y-,Z1}
{-$D-}   { Make it D+ for debug information }
{-$ObjExportAll On} { remove "-" to enable for C++ Builder 3.0!!!}
(************************************************************************
 * Compress.PAS V11.x - Delphi 6+, C++Builder 6+, Kylix 3               *
 *                                                                      *
 * Legal stuff:                                                         *
 *  Copyright(c) 1995-2014, Baskin-Kerr Trust under license to Webcentre*
 *  ALL RIGHTS RESERVED                                                 *
 *                                                                      *
 * This source code is provided for your inspection and adaptation.     *
 * While you are free to on-sell executable applications which make use *
 * of the source (in whole or in part) you may NOT:                     *
 * a) Use this source or any part of it to make components              *
 *    (DCUs, DLLs, OBJs, OCXs etc) for resale                           *
 * b) On-sell this source code or any part of it as part of any product *
 *                                                                      *
 * If you do not accept these restrictions, please contact              *
 * Webcentre Ltd for a full and immediate refund, or to obtain a        *
 * redistribution license tailored to suit your requirements.           *
 * Email: peter@webcentre.co.nz  Web: www.webcentre.co.nz               *
 *                                                                      *
 * Other than that -- have fun!                                         *
 ************************************************************************)
(* V11.x Amendment history:
    Sep 2014 V11.0 - compiled for XE7

 V10.x Amendment history:
    May 2014 V10.3 - compiled for XE6
    Oct 2013: V10.2 - compiled for XE5
    May 2013: V10.1 - small fix to restore compilability for non-Unicode, e.g. Delphi 7, XE4 compiled
    Sep 2012: V10.0 - compilable and tested in XE3, includes Win64

 V9.x Amendment history:
    Dec 2011: V9.4 - use RawByteString in string functions, LoadCompressedResource fix
    Nov 2011: V9.3 - CompressString fix, compiled for D2012/C2012 (Rad Studio XE2)
    Sep 2010: V9.2 - Compiled for D2011/C2011 (Rad Studio XE), no code changes
    Aug 2009: V9.1 - Compiled for D2010/C2010 (Rad studio 2010), no code changes
    Nov 2008: V9.0 - Added D2009 support. NB: filenames in archives are AnsiString (1-byte text)

V8.x Amendment history:
    Feb 2007: V8.3
               - remove some compiler warnings, test under D2007
      ...8.3 mods August 2006 (provided as source only in 8.3, no DCU changes):
               - fix bad checksum/last bytes when using .Expand() on a file > 2GB
                 (incorrect TCompressHeaderExSize value)
               - fail if empty-but-encrypted data is decrypted with wrong key
    Mar 2006: V8.2 -- NO source changes, just build/added Delphi 2006 support files
    Feb 2005: V8.1 -- fix bug in handling of Key property which made v8.0 incompatible with earlier versions
    Dec 2004: V8.0
     -- added support for files and archives bigger than 2 GBytes
       (FullSize and CompressedSize for TCompressHeader and TCompressFileHeader updated to Int64s)
     -- Delphi8/Delphi 2005 support added
     -- D3/D4/D5/BCB3/BCB4/BCB5/K1/K2 - no longer officially supported
     -- all pointer types replaced for .NET managed code

 V7.x amendment history:
   September 2002: V7.0 -- Delphi 7/Kylix 3 support added

*)

{ Disable any of the defines below to make a somewhat smaller component (see Help)
   -- at the cost of losing the particular feature noted.
   Note that the LZH compression routines are twice as large as the expansion ones...
 }
{$DEFINE LZH1_EXPAND}   { support expansion of colzh compressed files }
{$DEFINE LZH1_COMPRESS} { support compression of files with colzh }

{$DEFINE LZH5_EXPAND}   { support expansion of colzh5 compressed files (better than LZH) }
{$DEFINE LZH5_COMPRESS} { support compression of files with colzh5 }

 { These RLE routines are very small -- hardly worth the effort of disabling... }
{$DEFINE RLE_EXPAND}   { support expansion of RLE compressed files }
{$DEFINE RLE_COMPRESS} { support compression of files with RLE }

{ Don't change these bits: }
{$IFDEF LZH1_EXPAND}  { we want LZH1 expand enabled }
  {$DEFINE LZH1_COMMON}  { then we need common data/routines }
{$ENDIF}
{$IFDEF LZH1_COMPRESS}  { we want LZH1 compress enabled }
  {$DEFINE LZH1_COMMON}
{$ENDIF}
{$IFDEF LZH5_EXPAND}  { we want LZH5 expand enabled }
  {$DEFINE LZH5_COMMON}  { then we need common data/routines }
{$ENDIF}
{$IFDEF LZH5_COMPRESS}  { we want LZH5 compress enabled }
  {$DEFINE LZH5_COMMON}
{$ENDIF}

{ the two debugging defines below should be left disabled. }
{--$DEFINE COMPARE_ON_EXPAND}
{--$DEFINE COMPARE_ON_COMPRESS}
{$IFDEF COMPARE_ON_EXPAND}
  {$DEFINE COMPARE_BYTES}
{$ENDIF}
{$IFDEF COMPARE_ON_COMPRESS} { crude 'or' }
  {$DEFINE COMPARE_BYTES}
{$ENDIF}

{$IFDEF CLR}
{$R 'TCompress.bmp'}
{$ENDIF}

unit Compress;

interface

uses
{$IFDEF Linux}
  Qdialogs, Qforms,
  Types, Libc, IdGlobal,
{$ELSE}
  {$IFDEF CLR}
  WinUtils,
  {$ENDIF}
  Windows,
{$ENDIF}

  Variants,
  SysUtils, Classes;

{$IFDEF LZH1_COMMON} { stuff common to both compress and expand }
const constN  = 4096;                 { Buffer }
      TNIL    = 4096;                 { nil in tree terms }
      constF  = 60;                   { lookahead buffer }
      THRESHOLD = 2;

      N_CHAR   = (256 - THRESHOLD + constF); { 256-THRESHOLD+F, code : 0 .. N_CHAR-1 }
      constT   = N_CHAR * 2 - 1;             { size of table }
      constR   = constT - 1;                 { root position }
      MAX_FREQ = $8000;	                     { tree update timing from frequency }


type textbuffer = Array[0..constN + constF-1] of AnsiChar;
     lsons      = Array[0..constN] of Smallint;
     rsons      = Array[0..constN + 1 + constN] of Smallint;
     dads       = Array[0..constN + 1] of Smallint;
     samebuf    = Array[0..constN + 1] of AnsiChar;

     freqs      = Array[0..constT] of word;
     prnts      = Array[0..constT+N_CHAR] of smallint;
     sons       = Array[0..constT] of smallint;
{$ENDIF} { LZH1_COMMON }

{$IFDEF LZH5_COMMON} { stuff common to both compress and expand }
const MaxMatch = 256;
      DictionaryBits = 13;
      DictionarySize = 1 shl DictionaryBits;
      Hash1 = DictionaryBits - 9;
      Hash2 = DictionarySize * 2;
      CharMax = 255; { (1<<sizeof(char))*8 -1 }
      WordMax = 65535;
      WordBits = 16;
      SmallintMax = 32767;
      SmallintMin = SmallintMax - WordMax; { i.e. -32768 }
      max_hash_val = 3*DictionarySize + (DictionarySize div 512 + 1)*CharMax;
      BufferSize = 16*1024; { used encode only }
      Threshold5 = 3;
      constNC = CharMax + MaxMatch + 2 - Threshold5;
      CBIT = 9;
      PBIT = 4;         { smallest int s.t. (1<<PBIT) > NP }
      TBIT = 5;         { as above but > NT }
      constNP  =  13+1; { Max_DictionaryBits + 1 }
      constNT  = 16+3;  { sizeof(Smallint)*8 + 3 }
      constNPT = $80;
      NIL5 = 0;

type node = Smallint;
     textbuffer5      = Array[0..(DictionarySize*2+MaxMatch)] of AnsiChar;
     levelbuffer      = Array[0..(DictionarySize+CharMax+1)] of AnsiChar;
     childcountbuffer = Array[0..(DictionarySize+CharMax+1)] of AnsiChar;
     positionbuffer   = Array[0..(DictionarySize+CharMax+1)] of node;
     parentbuffer     = Array[0..(DictionarySize*2)] of node;
     prevbuffer       = Array[0..(DictionarySize*2)] of node;
     nextbuffer       = Array[0..(max_hash_val + 1)] of node;
     buffer           = Array[0..BufferSize] of AnsiChar; { used for encode only }
     childBuffer      = Array[0..2*constNC-1] of Word;
     chartablebuffer  = Array[0..4096] of Word;
     charcodebuffer   = Array[0..constNC] of Word;
     charlenbuffer    = Array[0..constNC] of AnsiChar;
     positiontablebuffer = Array[0..256] of Word;
     positioncodebuffer  = Array[0..constNPT] of Word;
{$ENDIF} { LZH5_COMMON }

type
  ELZHBadDecodeTable = Class(Exception); { public V3.51 }
  EUnableToCompress = Class(Exception);  { public V3.51 }

  EInsufficientDiskSpace = Class(Exception);
  EUnrecognizedCompressionMethod = Class(Exception);
  EInvalidHeaderArchiveType = Class(Exception);
  EInvalidHeader = Class(Exception);
  EBadChecksum = Class(Exception);
  EInvalidKey = Class(Exception);

  TCompressArchiveType = (caSingle, caMulti);
  TCompressionMethod = ( coNone, coRLE, coLZH, coCustom, coLZH5 );
  TCProcessMode = ( cmCompress, cmExpand, cmDelete ); { for ProcessStreams/OnCheckFile }

{$ALIGN OFF}  
  TCompressHeader =
      Record
        ComId: array[0..4] of AnsiChar; { Always SPIS + chr(26) }
        ComMethodId: array[0..2] of AnsiChar; { RLE, LZH, LH5 etc. }
        Fullsize: Int64;      { V8.0 upgraded from Longint }
                              { uncompressed size, in bytes, excluding header }
        ArchiveType: TCompressArchiveType; { caSingle, caMulti }
        Checksum: Longint;    { CRC or checksum -- not used in multi-file archive }
        Locked: Longint;      { >0 means file was saved with a key V2.5 }
      end;

  TCompressedFileHeader =
      Record
        FileNameLength: Smallint; // v9 - to be clear, this is size in BYTES
        DateTime: Longint;
        Attributes: Smallint;
        Fullsize: Int64;       { V8.0 upgraded from Longint }
        CompressedSize: Int64; { V8.0 upgraded from Longint }
        CompressedMode: TCompressionMethod;
        Checksum: Longint;   { CRC or checksum }
        Locked: Longint;     { 1 means file was Locked by TCompress 3.0 or earlier,
                               2 means file was encrypted by TCompress 3.5 or later }
      end;
{$ALIGN ON}

  TCompressedFileInfo = Class(TObject){ for a compressed file directory listing }
  public
        Datetime: Longint;
        Attributes: Smallint;
        Fullsize: Int64;        { V8.0 upgraded from Longint }
        CompressedSize: Int64;  { V8.0 upgraded from Longint }
        CompressedMode: TCompressionMethod;
        Checksum: Longint;
        Position: Int64;   { offset of start of HEADER in the file }
                           { V8.0 upgraded from Longint }
        Locked: Boolean;   { was file saved with a key? }
  end;

  TFilenameCheckEvent = procedure (var filepath: String;
                                    mode: TCProcessMode) of Object;
  TCompressEvent = procedure(dest,source: Tstream;
                    var CompressID : AnsiString; var Outputsize: Int64;  { V8.0 upgraded from Longint }
                                          var checksum: Longint) of Object;  // v9, CompressID was string
  TExpandEvent = procedure(dest,source: Tstream; Sourcesize, Destsize: Int64;  { V8.0 upgraded from Longint }
                           CompressID: AnsiString; var checksum: Longint) of Object;  // v9, CompressID was string
  TRecognizeEvent = procedure(CompressID: AnsiString;
                                 var recognized: Boolean) of Object;  // v9, CompressID was string

  TShowProgressEvent = procedure(var PercentageDone: Longint) of Object;  { V8.0 keep it as Longint for backward compatibility with existing projects }

  TCompThreadSafety = class
    AtStart, InputEOF, inRepeat: Boolean;
    inBuffer, outBuffer: array of Byte;
    inBuffPos, outBuffPos: Integer;
    source, dest: TStream;
    lastch: AnsiChar;
    DupCount : Integer;
    InChecksum, Outchecksum: Longint;
    Readsize, lChunk, BytesIn, BytesOut: Int64;  { V8.0 upgraded from Longint }
    EncryptingIn, EncryptingOut: Boolean; { V3.5 }
    EncryptCodes: Array[0..3] of Byte;
    EncryptCounter: Smallint;
    MaxNumberOfBytesToWrite: Int64;  { V8.0 upgraded from Longint }
    GlobalFOnShowProgress: TShowProgressEvent;

{$IFDEF LZH1_COMMON} { vars common to both compress and expand }
    text_buf: textbuffer;
    lson:   lsons;
    rson:   rsons;
    dad:    dads;
    same:   samebuf;
    freq:   freqs;  	  { frequency table }
    prnt:   prnts;      { parent nodes }
    son:    sons;       { points to son nodes (son[i],son[i+1]) }
    putbuf: Word;
    putlen: AnsiChar;
    getbuf: Word;
    getlen: AnsiChar;
    match_position, match_length: Word; { really only LZH1_COMPRESS but what the heck }
{$ENDIF} { LZH1_COMMON }

{$IFDEF LZH5_COMMON} { vars common to both compress and expand }
    text: textbuffer5;
    level: levelbuffer;
    childcount: childcountbuffer;
    position: positionbuffer;
    parent: parentbuffer;
    prev: prevbuffer;
    buf: buffer;                    // encode only?
    subBitbuf, Bitcount: AnsiChar;  // mini ints really...
    Blocksize, bitbuf: Word;
    left: childbuffer;              // all these arrays oversize by 1
    right: childbuffer;
    charLen: charlenbuffer;
    position_len: array[0..constNPT] of AnsiChar;
    char_freq: childbuffer;
    char_table: chartablebuffer;
    char_code: charcodebuffer;
    position_freq: array[0..2 * constNP - 1] of Word;
    position_table: positiontablebuffer;
    position_code: positioncodebuffer;
    t_freq: array[0..2 * constNT - 1] of Word;
    pos, MatchPosition, avail: node;
    next: nextbuffer;
    depth: AnsiChar;
{$IFDEF COMPARE_BYTES}
    compareFile: TFilestream; { used for debugging only }
{$ENDIF}
{$ENDIF} { LZH5_COMMON }

{$IFDEF LZH5_COMPRESS} { compress vars only }
    output_pos, output_mask, cpos: Word;
    remainder, MatchLength: Smallint;
{$ENDIF} { LZH5_COMPRESS }

    function GetChar: AnsiChar;
    function Encrypt(ch: AnsiChar): AnsiChar;

    procedure PutChar(ch: AnsiChar);
    procedure CopyWithCRC;

{$IFDEF RLE_COMPRESS}
    procedure emit(count: Integer; ch: AnsiChar);
    procedure CompressRLE;
{$ENDIF}

{$IFDEF RLE_EXPAND}
    procedure ExpandRLE;
{$ENDIF}

{$IFDEF LZH1_COMMON} { routines common to both compress and expand }
    procedure StartHuff;
    procedure EndHuff;

    procedure reconst;
    procedure update (c: Word);
{$ENDIF} { LZH1_COMMON }

{$IFDEF LZH1_COMPRESS}
    procedure InitTree;
    procedure InsertNode(r: Smallint);
    procedure link (n, p, q: Smallint);
    procedure linknode (p, q, r: smallint);
    procedure DeleteNode(p: Smallint);

    procedure Putcode (L: Smallint; c: Word);
    procedure EncodeChar (c:word);
    procedure EncodePosition(c:Word);
    procedure EncodeEnd;
    procedure CompressLZH;
{$ENDIF}

{$IFDEF LZH1_EXPAND}
    function GetBit: Smallint;
    function GetByte: Word;
    function GetNBits(n: Word): Smallint;
    function DecodeChar: Smallint;
    function DecodePosition: Smallint;
    procedure ExpandLZH(textsize: Int64);  { V8.0 upgraded from Longint }
{$ENDIF}

{$IFDEF LZH5_COMMON} { routines common to both compress and expand }
    procedure StartHuff5(Compress: Boolean); { init bit get/put vars }
    procedure EndHuff5;
{$IFDEF COMPARE_BYTES}
    procedure chkchar(c: smallint); { used for debugging only }
{$ENDIF}
{$ENDIF} { LZH5_COMMON }

{$IFDEF LZH5_EXPAND}
    procedure buildtable(nchar: Smallint; bitlen: Array of AnsiChar;
                            tablebits: Smallint; var table: Array of Word);
    procedure fillbuf(n: AnsiChar);
    function getbits(n: AnsiChar): Word;
    procedure read_position_len(nn, nbit, i_special: Smallint);
    procedure read_charLen;
    function DecodeChar5: Word;
    function DecodePosition5: Word;
    procedure ExpandLZH5(destsize: Int64);  { V8.0 upgraded from Longint }
{$ENDIF}

{$IFDEF LZH5_COMPRESS}
    function buildtree(nparm: Smallint; var freqparm: Array of Word; var lenparm: Array of AnsiChar;
                       var codeparm: Array of Word): Smallint;
    procedure count_t_freq;
    procedure initSlide;
    function child(q: node; c: AnsiChar): node;
    procedure makechild(q: node; c: AnsiChar; r: node);
    procedure split(old: node);
    procedure insert_node;
    procedure delete_node;

    procedure putcode5(n: AnsiChar; x: Word);
    procedure putbits(n: AnsiChar;x: Word);
    procedure write_position_len(n, nbit, i_special: Smallint);
    procedure write_charLen;
    procedure encode_char(c: Smallint);
    procedure encode_position(p: Word);
    procedure send_block;
    procedure OutputCodes(c, p: Word);
    procedure get_next_match;
    procedure CompressLZH5;
{$ENDIF}
  end;

  TCompress = class(TComponent)
  private
    Global : TCompThreadSafety;
    FIgnoreCase: Boolean; { Linux-only case matching control }
    FRegName: String;
    FRegNumber: Longint;
    FCompressedToPercentage: Integer;
    FCompressionTime: DWORD;
    FOnCompress: TCompressEvent;
    FOnExpand: TExpandEvent;
    FOnRecognize: TRecognizeEvent;
    FOnCheckFile: TFilenameCheckEvent;
    FOnShowProgress: TShowProgressEvent;

    FTargetPath: string;
    FMakeDirectories: Boolean;
    FKeyNum: Longint;
    FExceptionOnFileError: Boolean;
    FCheckSpaceBeforeExpand: Boolean; { V3.05 }

    FMustCompressByAtLeast: Int64; { V3.5 } { V8.0 upgraded from Longint }

    hFileInstance : THandle; // v9.4 for LoadCompressedResource handling


{$IFDEF UNICODE}
    function ProcessString(const str: rawbytestring; compressionMethod: TCompressionMethod;
                           mode: TCProcessMode): rawbytestring; // rawbytestring v9.4
{$ELSE}
    function ProcessString(const str: string; compressionMethod: TCompressionMethod;
                           mode: TCProcessMode): string; { V3.5 }
{$ENDIF}
    function CheckDiskSpace(filename: String): Int64; { class member, V3.05 } { V8.0 upgraded from Longint }
    procedure SetTargetPath(const path:string);
    function GetCompressedPercentage: Integer;
    function ProcessStreams(outstream, instream: TStream; Sourcesize, destSize: Int64;  { V8.0 upgraded from Longint }
                            CompressionMethod: TCompressionMethod; var checksum: longint;
                            mode: TCProcessMode; Encrypted: Boolean): Int64; { encrypted V3.5 }
                                               { V8.0 upgraded from Longint }
    function MatchFileName(const Filename, FilenameMask: string): Boolean;
    function FileInList(whichFiles: TStringList;filename: string): Boolean;

  protected
    { Protected declarations }
    procedure AppendFilesExcept(const destfile,fromfile: String; notfiles: TStringlist);
    procedure SetHeader(var hdr: TCompressHeader; cID: AnsiString; // v9 string --> AnsiString
                          aType: TCompressArchiveType; size: Int64);  { V8.0 upgraded from Longint }
    function GetFileHeader(Stream: TStream;var Fhdr: TCompressedFileHeader; var Filename: String): Boolean;
    procedure PutFileHeader(Stream: TStream;var Fhdr: TCompressedFileHeader; const Filename: String);
    procedure CheckHeader(compressedFile: TStream;var hdr: TCompressHeader; var cID: AnsiString); // v9 string --> AnsiString

    // The following 4 functions were added in V8.0; for compatibility, they all return
    // LongInts (Read and Write methods of all TStream descendants return LongInts)
    function ReadHeader(Stream: TStream; var hdr: TCompressHeader): LongInt;             { V8.0 }
    function WriteHeader(Stream: TStream; var hdr: TCompressHeader): LongInt;            { V8.0 }
    function ReadFileHeader(Stream: TStream; var Fhdr: TCompressedFileHeader): LongInt;  { V8.0 }
    function WriteFileHeader(Stream: TStream; var Fhdr: TCompressedFileHeader): LongInt; { V8.0 }

    // Our own SizeOf functions for archive header and file header
    function SizeOfHeader(hdr: TCompressHeader): Integer; overload;
    function SizeOfFileHeader(Fhdr: TCompressedFileHeader): Integer;

    procedure Move(source: string; var dest: array of AnsiChar); overload;
    procedure Move(source: array of AnsiChar; var dest: string); overload;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function DiskFree(Drive:byte): Int64; { upgraded to Int64 in V8.0; added v4.0 - made public v5.0}
{$IFDEF UNICODE}
    function CompressString(const unCompressedString: rawbytestring; compressionMethod: TCompressionMethod): rawbytestring;  // rawbytestring v9.4
    function ExpandString(const CompressedString: rawbytestring): rawbytestring;  // rawbytestring v9.4
{$ELSE}
    function CompressString(const unCompressedString: string; compressionMethod: TCompressionMethod): string;  // rawbytestring v9.4
    function ExpandString(const CompressedString: string): string;  // rawbytestring v9.4
{$ENDIF}

    function DoCompress(compressedStream, uncompressedStream: Tstream;
       compressionMethod: TCompressionMethod; var CompressID : AnsiString;  // v9, CompressID was string
                                              var checksum: Longint): Int64;  { V8.0 upgraded from Longint }
    procedure DoExpand(expandedStream, unexpandedStream: Tstream;
                       compressedsize, fullsize: Int64;  { V8.0 upgraded from Longint }
                       checksum: Longint;
                       compressionMethod: TCompressionMethod; const CompressID: AnsiString;  // v9, CompressID was string
                       FLocked: Longint); { FLocked now Longint V3.5 }
    function Recognize(const cID: AnsiString) : TCompressionMethod; // v9 string --> AnsiString
    procedure Compress(compressedStream, uncompressedStream: Tstream;
                                         compressionMethod: TCompressionMethod);
    procedure Expand(expandedStream,unexpandedStream: Tstream);

    procedure CompressStreamToArchiveStream(compressedStream,uncompressedStream: TStream; { V3.0 }
         const Filename: string; FHdr: TCompressedFileHeader; var cID: AnsiString; // v9, cID was string
         CompressionMode: TCompressionMethod);

    procedure CompressStreamToArchive(arcFile: String; uncompressedStream: TStream;
             const fileName: string; CompressionMode: TCompressionMethod); { V3.0 }

    procedure ExpandStreamFromArchive(arcFile: String; uncompressedStream: TStream; fileName: string); { V3.0 }

    function CompressFilesToStream(dest: TStream; var whichfiles: TStringList;
                                   CompressionMode: TCompressionMethod): Int64;  { V8.0 upgraded from Longint }
    procedure CompressFiles(const arcFile: String; whichfiles: TStringList;
                                   CompressionMode: TCompressionMethod);
    procedure CompressFile(const arcfile:String; var fromFile: String; CompressionMode: TCompressionMethod);
    procedure DeleteFiles(const arcFile: String; whichfiles: TStringList);
    procedure ExpandFile(var toFile:String;const arcfile: String);
    procedure ExpandFiles(const arcfile:String;whichfiles: TStringList);
    procedure ExpandFilesFromStream(compressedstream:TStream;whichfiles: TStringList);
    procedure ScanCompressedStream(compressedstream: TStream; var Finfo: TStringList);
    procedure ScanCompressedFile(const arcfile: String; var Finfo: TStringList);
    property CompressedPercentage: Integer read GetCompressedPercentage;
    property CompressionTime: DWORD read FCompressionTime;
    function LoadCompressedResource(ResourceName, DLLName: string): TStream;
    function LoadExpandedResource(ResourceName, DLLName: string): TStream;
    procedure ReleaseCompressedResource;
    procedure FreeFileList(Finfo: TStringList);
    procedure GetAllFilesInDir(list: TStringList; dirname: string; const match: string; Anything: Boolean); { new in V3.0 }
    procedure GetMatchingFiles(list: TStringList; const matchname: string); { V3.0 }

  published
{$IFDEF Linux}
    property IgnoreCase: Boolean read FIgnoreCase write FIgnoreCase default False; { Linux-only case matching control }
{$ENDIF}
    property RegName : string read FRegName write FRegName;
    property RegNumber : Longint read FRegNumber write FRegNumber;
    property OnCompress : TCompressEvent read FOnCompress write FOnCompress;
    property OnExpand : TExpandEvent read FOnExpand write FOnExpand;
    property OnCheckFile : TFilenameCheckEvent read FOnCheckFile write FOnCheckFile;
    property OnRecognize : TRecognizeEvent read FOnRecognize write FOnRecognize;
    property OnShowProgress : TShowProgressEvent read FOnShowProgress write FOnShowProgress;

    property TargetPath : String read FTargetPath write SetTargetPath;
    property MakeDirectories : Boolean read FMakeDirectories write FMakeDirectories;
    property ExceptionOnFileError: Boolean read FExceptionOnFileError write FExceptionOnFileError;
    property Key: Longint read FKeyNum write FKeyNum;
    property CheckSpaceBeforeExpand: Boolean read FCheckSpaceBeforeExpand
                             write FCheckSpaceBeforeExpand; { V3.05 }
  end;

const CompressNoMoreFlag='*';
      CompressSkipFlag='';
      ConstMaxINT64     = High(Int64);  { added V8.0 }

procedure Register;

implementation

const
{$IFDEF Linux}
    DLGVal      : Cardinal = $00000000 + 00000040 + $00000000;
    Slash       : Char     = '/';
    ConfirmVal  : Cardinal = $00000000 + $00000020 + $00000003;
{$ELSE}
    DLGVal= MB_APPLMODAL+MB_ICONINFORMATION+MB_OK;
    Slash       : Char     = '\';
{$ENDIF}

    ChunkSize = 8192; { copy 8K chunks }
    CompressedFileHeaderID = 'SPIS'+chr(26);
    CustomCompressionID='CUS';
    NoCompressionID='NON';
    RLECompressionID='RLE';
    LZHCompressionID='LZH';
    LZH5CompressionID = 'LH5';
    TempExtension='.$$$';
    LockedFlag = 1;  { V3.0 or earlier }
    EncryptFlag = 2; { V3.5 or later }

    BigFileFlag = $80;     { V8.0 bit 7 of Locked is on for files > 2 GB }
    BigFileFlagMask = $7f; { V8.0 }

  { WARNING: if any items in the header/fileHeader changes it's size the next
    four constants must be also uptaded }
    TCompressHeaderSize  = 21;   { V8.0 - TCompress archive Header - standard header size}
    TCompressHeaderExSize= 25;   { V8.3 - was incorrectly 29 in earlier versions - 'big-files' header size}
    TCompFileHeaderSize  = 25;   { V8.0 - File Header in TCompress archive - standard header size}
    TCompFileHeaderExSize= 33;   { V8.0 - File Header in TCompress archive - 'big-files' header size}

    CompressionCodes: array[TCompressionMethod] of AnsiString = ( // v9 was string
           NoCompressionID,     { coNone }
           RLECompressionID,    { coRLE }
           LZHCompressionID,    { coLZH }
           CustomCompressionID, { coCustom }
           LZH5CompressionID);  { coLZH5 }

    RLEescapechar: AnsiChar = AnsiChar(148); { thus 65 148 37 is 37 A's, and 148 00 is ONE 148 }

    TCOMPRESS_RESOURCE_TYPE = 'TCOMPRESS';

{$IFDEF LINUX}  { Utility routines not currently defined to exact same spec in Kylix }
function MessageBox(iHandle : integer; MessageStr, CaptionStr : PChar; MessageType : Cardinal): Integer;
begin
  result := 0;
  if MessageType = ConfirmVal then begin
    result := MessageDLG(String(CaptionStr), String(MessageStr), mtConfirmation,[mbYes, mbNo, mbCancel], 0, mbYes);
  end
  else if MessageType = DLGVal then begin
    result := MessageDLG(String(CaptionStr), String(MessageStr), mtInformation, [mbOK], 0, mbOK);
  end;
end;

function FileGetAttr(const FileName: string): Integer;
var searchRec: TSearchRec;
begin
  result := -1;
  if FindFirst(FileName,faAnyFile,SearchRec)=0 then
    result := SearchRec.attr;
end;
{$ENDIF} { Linux}


function CalcPercentage(numerator, denominator: Int64): Integer; { new 3.01 to avoid o'flo probs }
begin                                        { V8.0 parameters upgraded from Longint }
  if denominator > 100000 then { we can stand dividing it some! }
    result := Integer(numerator div (denominator div 100))
  else
    result := Integer(100*numerator div denominator); { conservative, but overflows if numerator > 20MB }
end;

constructor TCompress.Create(AOwner:TComponent);
begin
  inherited create(AOwner);
  global := TCompThreadSafety.create;
  hFileInstance := 0;
end;

destructor TCompress.Destroy;
begin
  ReleaseCompressedResource; // if there is one
  global.free;
  inherited destroy;
end;

{$IFDEF UNICODE}
function TCompress.ProcessString(const str: rawbytestring; compressionMethod: TCompressionMethod; // v9.4 rawbytestring
          mode: TCProcessMode): rawbytestring;
{$ELSE}
function TCompress.ProcessString(const str: string; compressionMethod: TCompressionMethod;
          mode: TCProcessMode): string; // V10.1 - removed erroneous rawbytestring
{$ENDIF}
var outMS, inMS: TMemoryStream;
{$IFDEF CLR}
    tmpStrBuff: array of Byte;
{$ENDIF}
begin
   outMS := TMemoryStream.create;
   inMS := TMemoryStream.create;
   try
{$IFDEF CLR}
     inMS.write(BytesOf(str),Length(str) * sizeof(str[1])); // v9 add sizeof for WideStrings
{$ELSE}
     inMS.write(str[1],Length(str) * sizeof(str[1])); // v9 add sizeof for Widestrings
{$ENDIF}
     inMS.seek(0,soBeginning);
     if mode = cmCompress then
       Compress(outMS,inMS,compressionMethod)
     else
       Expand(outMS,inMS);

     with outMS do
     begin
       setLength(result,trunc(0.9 + outMS.size / sizeof(str[1]))); // v9.3 0.9 factor ensures stray string bytes are allowed for in widestrings
       seek(0,soBeginning);
{$IFDEF CLR}
       setLength(tmpStrBuff, outMS.Size);
       read(tmpStrBuff,outMS.size);
       result := StringOf(tmpStrBuff);
{$ELSE}
       read(result[1],outMS.size);
{$ENDIF}
     end;
   finally
     outMS.free;
     inMS.free;
   end;
end;

{$IFDEF UNICODE}
function TCompress.CompressString(const unCompressedString: rawbytestring;
             compressionMethod: TCompressionMethod): rawbytestring;  // v9.4 rawbytestring
{$ELSE}
function TCompress.CompressString(const unCompressedString: string;
             compressionMethod: TCompressionMethod): string;
{$ENDIF}

begin
   result := ProcessString(unCompressedString,compressionMethod,cmCompress);
end;

{$IFDEF UNICODE}
function TCompress.ExpandString(const CompressedString: rawbytestring): rawbytestring;   // v9.4 rawbytestring
{$ELSE}
function TCompress.ExpandString(const CompressedString: string): string;
{$ENDIF}
begin
   result := ProcessString(CompressedString,coNone,cmExpand);
end;


{ Functions to support /S processing }

{ NB: WildMatch is based on the "MatchString" routine from Michael
  Ax's TPack component set. Adapted with kind permission: ax@href.com }
function WildMatch(Const Source, Match: string): Boolean;
{matches 'Hellothere' to 'He???th*' -- Matches even if Source longer than match!}
var
  i,n:integer;
  hasStar: Boolean;
begin
  Result:= (Match='*') or ((Match='') and (Source=''));
  n:=length(Match);          {min match length}
  i:=pred(pos('*',Match));   {use '*' as cutoff}
  hasStar := i > 0;

  if hasStar and (i<n) then      {did we get one}
    n:=i;                      {assign the earlier}
  if length(source)<n then     {is the input long enough}
    exit;                      {nope. exit with result set}
  for i:= 1 to n do            {loop through match-string}
    if Result then
      exit
    else
      if (Match[i]<>'?') and (Match[i]<>Source[i]) then
        break
      else
        Result:= (i=n) and
         (hasStar or (i=Length(Source))); { NO implicit * at end of match this way... }
{        Result:=i=n; {done?}
  {returns true if the source matches the pattern. trailing '*' implied in pattern!!}
end;

{ Changed in 6.1 -- bug fix concerning '.' spec below PJH 19/10/2001 }
{Changed in 6.0 - now deals with directory names with '.' in them
    and supports Linux-only IgnoreCase property }
function TCompress.MatchFileName(const Filename, FilenameMask: string): Boolean;
var lpf,rpf,lpm,rpm: string;
begin

{$IFDEF Linux}
 if FIgnoreCase then
 begin
{$ENDIF}

     { this logic is ALWAYS true for Windows: }
    rpm := ExtractFileExt(AnsiLowerCase(FileNameMask));
    lpm := AnsiLowerCase(copy(FilenameMask,1,(length(FilenameMask)-Length(rpm))));
    if rpm='' then rpm :='*';
    rpf := ExtractFileExt(AnsiLowerCase(FileName));
    lpf :=AnsiLowerCase(copy(Filename,1,(length(Filename)-Length(rpf))));
{ Moved AFTER line above in 6.1 19/10/2001 }
    if rpf='' then rpf:='.'; { V6.0 so ABC matches *.* }
    result := WildMatch(lpf,lpm) and WildMatch(rpf,rpm);

{$IFDEF Linux}
 end else { make it case SENSITIVE -- Linux-only spec }
 begin
    rpm := ExtractFileExt(FileNameMask);
    lpm := copy(FilenameMask,1,(length(FilenameMask)-Length(rpm)));
    if rpm='' then rpm :='*';
    rpf := ExtractFileExt(FileName);
    if rpf='' then rpf:='.'; { V6.0 so ABC matches *.* }
    lpf :=copy(Filename,1,(length(Filename)-Length(rpf)));
    result := WildMatch(lpf,lpm) and WildMatch(rpf,rpm);
 end; { if FIgnoreCase }
{$ENDIF} { Linux }

end;

{ Below are two new routines introduced in V3.0 (/S supported added V3.52) }

procedure TCompress.GetAllFilesInDir(list: TStringList; dirname: string; const match: string; Anything: Boolean);
var Searchrec: TSearchRec; { used for FindFirst/FindNext }
    FoundOne : Integer;
    Searchfor: string;
    recurselist: TStringlist;
    recurse: Boolean;
begin
     if list=nil then exit;
     list.clear;
     if Length(dirname) > 0 then { not "current" }
        if dirname[Length(dirname)]<>slash then
            dirname:=dirname+slash;  { need \ for legal paths }

     if (Length(Match) > 2) and (CompareText(copy(Match,Length(Match)-1,2),'/S')=0) then
     begin { V3.52 }
        SearchFor :=copy(Match,1,Length(Match)-2); { back to normal case }
        recurse := True;            { but...}
     end else
     begin
        recurse := False;
        SearchFor := match; { v2.5a -- was dirname+match }
     end;
{$R-} { RangeChecks off, in case on at this point...}
     FoundOne := FindFirst(dirname+'*.*',faAnyFile,SearchRec);
     while FoundOne=0 do
     begin
       with SearchRec do
       if (Copy(Name,1,1)<>'.') then { not a . or .. reference }
       begin
        if recurse and ((Attr and faDirectory)<>0) then { recursing? v3.52 }
        begin { get all files from lower level(s) }
           recurseList := TStringList.create;
           try
              GetAllFilesInDir(recurselist,dirname+name,match,anything);
              list.addstrings(recurseList);
           finally
             recurseList.free;
           end;
        end;
           if
           (anything or (                    {changed V5.0}
           (attr and faDirectory <= 0) and  {These changes done to allow the}
           (attr and faHidden <=0) and      {detection of Archive files that}
           (attr and faSysFile <=0) and     {were in WinNT compressed directories}
           (attr and faVolumeID <=0) and        {rather than test for solely faArchive }
           MatchFileName(Name,SearchFor)
           )) then {also now picks up Read only files}
              list.add(dirname+Name);
       end; { with }
       FoundOne := FindNext(SearchRec); { awkward, but that's how First/Next work... }
     end;
     SysUtils.FindClose(SearchRec);
end;

{ Gets a list of all regular files (excludes System, Hidden, Directory AND ReadOnly)
  which MATCH the name specified (can be wild, can't be just a dirname)
  -- returns in LOWER case, full path as per matchname passed in
}
procedure TCompress.GetMatchingFiles(list: TStringList; const matchname: string);
var dirpath: string;
begin
     if Length(matchname)=0 then exit;
     dirPath := ExtractFilePath(matchname);
     GetAllFilesInDir(list,dirPath,ExtractFileName(matchname),False); { regular files only }
end;


procedure TCompress.SetTargetPath(const path:string);
begin
  FTargetpath := path;
  if path<>'' then
     if Copy(path,Length(path),1)<>slash then
        FTargetpath := path+slash;
end;

{ Call this function to empty the list populated by a
  call to ScanCompressedFile or ScanCompressedStream
  Clears the stringlist as well as the freeing the objects in it.
}
procedure TCompress.FreeFileList(Finfo: TStringList);
var count: Integer;
begin
  if Finfo<> nil then
  begin
    for count:= 0 to pred(Finfo.count) do
     Finfo.objects[count].free; { get rid of these (if any)... }
    Finfo.clear; { changed in V3.0 -- in v2.5 this was a FREE! }
  end; { begin/end added v3.05 }
end;

{ This function can be called to load a resource from either the current EXE or,
  if DLLName is not empty, from a named DLL. Returns a pointer to either a
  TMemoryStream (Delphi 1) or a TResourceStream (Delphi 2) containing the
  resource required.  Resources are assumed to be of the type 'TCOMPRESS'.
}
function TCompress.LoadCompressedResource(ResourceName, DLLName: string): TStream;
// var  hFileInstance :  THandle;  // v9.4 now using global
begin
  if DLLName <> '' then
  begin
{$IFDEF CLR}
    hFileInstance := LoadLibrary(DLLName);   { Load DLL }
    if hFileInstance = 0 then { oops }
{$ELSE}
    hFileInstance := LoadLibrary(PChar(DLLName));   { Load DLL }
    if hFileInstance = 0 then { oops }
{$ENDIF}
       raise(Exception.create('Could not load '+DLLName+' DLL'));
  end else
    hFileInstance := HInstance; { the handle to the current EXE }
  try
{$IFDEF CLR}
    result := TResourceStream.Create(hFileInstance,ResourceName,TCOMPRESS_RESOURCE_TYPE);
{$ELSE}
    result := TResourceStream.Create(hFileInstance,ResourceName,PChar(TCOMPRESS_RESOURCE_TYPE));
{$ENDIF}
  finally
//    if DLLName <> '' then { must free Library }  pre v9.4 OLD CODE - too soon for DLLs in later OSes
//      FreeLibrary(hFileInstance );     { Always Unload DLL }
      if DLLName = '' then   // v9.4 ensure cleared so no FreeLibrary call can happen
        hFileInstance := 0;
  end;
end;

procedure TCompress.ReleaseCompressedResource; // new in v9.4
begin
  if hFileInstance <> NULL then
    try
      FreeLibrary(hFileInstance);
      hFileInstance := 0;
    except on e:exception do // silent fail
      hFileInstance := 0;
    end;
end;


{ Tries to load AND expand compressed resource of the given name from the given DLL/EXE
 (or the current app if DLLName is blank). Raises exception if not found. }
function TCompress.LoadExpandedResource(ResourceName, DLLName: string): TStream;
var CompressedStream: TStream;
begin
     result := TMemoryStream.Create;
     CompressedStream :=LoadCompressedResource(ResourceName,DLLName);
     try
        if (CompressedStream=nil) or (CompressedStream.size=0) then
           raise(EResNotFound.Create('Could not load resource: '+ResourceName));
        Expand(result,CompressedStream); { got it }
        result.seek(0,soBeginning); { rewind! }
     finally
       CompressedStream.free;
       ReleaseCompressedResource; // v9.4
     end;
end;

function DirectoryExists(const Name: string): Boolean;
var
  Code: Integer;
begin
  Code := FileGetAttr(Name);  { v9, formerly had IFDEF CLR/ELSE Pchar logic, now redundant }
  Result := (Code <> -1) and ((faDirectory and Code) <> 0); { v6.0 was FILE_ATTRIBUTE_DIRECTORY }
end;

{$IFDEF NOLONGERNEEDED} // v9 - fully supported in Sysutils now
procedure ForceDirectories(Dir: string);
begin
  if Dir='' then exit;  { note: this test not in VCL code }
  if Dir[Length(Dir)] = slash then
    SetLength(Dir, Length(Dir)-1);
  if (Length(Dir) < 3) or DirectoryExists(Dir) then Exit;
  ForceDirectories(ExtractFilePath(Dir));
  MkDir(Dir);
end;
{$ENDIF}

procedure TCompress.CompressFile(const arcFile:String; var fromFile: String;
              CompressionMode: TCompressionMethod);
var s: TStringlist;
begin
  s:=TstringList.create;
  try
     s.add(fromfile);
     CompressFiles(arcFile,s,compressionMode);
     if s.count > 0 then
       fromfile:=s[0] { in case it changed }
     else
       fromFile :='';
  finally
     s.free;
  end;
end;

procedure TCompress.ExpandFile(var toFile: String;const arcfile: String);
var s: TStringlist;
begin
  s:=TstringList.create;
  try
     s.add(tofile);
     ExpandFiles(arcfile,s);
     if s.count > 0 then
       tofile:=s[0] { in case it changed }
     else
       toFile :='';
  finally
     s.free;
  end;
end;

function TCompress.DoCompress(compressedStream, uncompressedStream: Tstream;
        compressionMethod: TCompressionMethod; var CompressID : AnsiString;  // v9, CompressID was string
                                             var checksum: Longint): Int64;  { V8.0 upgraded from Longint }
begin
  FCompressionTime := GetTickCount;
  Result := 0; { Will only be in catastrophic failures? }
  if CompressionMethod=coCustom then
  begin
    checksum := 0;
    if assigned(FOnCompress) then
      FOnCompress(compressedStream,unCompressedStream, CompressID ,Result, checksum);
  end else
    result := ProcessStreams(compressedStream,uncompressedStream,
                             uncompressedStream.Size-uncompressedStream.Position, { V8.0 - fix for CDBImage streams }
                             0,compressionMethod,checksum,cmCompress, False);
  Checksum := CheckSum+Key;
  FCompressionTime:=GetTickCount - FCompressionTime;
end;

procedure TCompress.SetHeader(var hdr: TCompressHeader; cID: AnsiString; // v9, cID was string
                           aType: TCompressArchiveType; size: Int64);  { V8.0 upgraded from Longint }
{$IFDEF CLR}
var tmpStr: String;
{$ENDIF}
begin
  with hdr do
  begin
{$IFDEF CLR}
    tmpStr := CompressedFileHeaderID;
    Move(tmpStr, ComID);                                     { terminator goes in too, but is safe here }
    Move(cID,ComMethodID);                                   { avoid terminator }
{$ELSE}
    strPLCopy(ComID, CompressedFileHeaderID, sizeof(ComID)); { terminator goes in too, but is safe here }
    System.Move(cID[1],ComMethodID,sizeof(ComMethodID));     { avoid terminator }
{$ENDIF}
    if (size > Int64(MaxLongInt)) and (aType = caMulti) then { V8.0 - for Multi file archives bigger   }
      Fullsize := 0                                          {        than 2 GB we ignore hdr.Fullsize }
    else
      Fullsize := size; { uncompressed size, in bytes }
    ArchiveType := aType;
    Checksum := 0;
    Locked := 0;
    if Key>0 then
      Locked := EncryptFlag; { V3.5 }
  end;
end;

function TCompress.ReadHeader(Stream: TStream; var hdr: TCompressHeader): LongInt;   { V8.0 }
var UpperLongIntOfSize: LongInt;
    TmpLongInt: LongInt;
    Idx: Integer;
    aByte: Byte;
begin
  Result := 0;
  for Idx := 0 to 4 do
    Result := Result + Stream.Read(hdr.ComId[Idx], SizeOf(hdr.ComId[Idx]));
  for Idx := 0 to 2 do
    Result := Result + Stream.Read(hdr.ComMethodId[Idx], SizeOf(hdr.ComMethodId[Idx]));
  Result := Result + Stream.Read(TmpLongInt, SizeOf(TmpLongInt));
  hdr.Fullsize := Int64(TmpLongInt);
  Result := Result + Stream.Read(aByte, SizeOf(aByte));
  hdr.ArchiveType := TCompressArchiveType(aByte);
  Result := Result + Stream.Read(hdr.Checksum, SizeOf(hdr.Checksum)) +
                     Stream.Read(hdr.Locked, SizeOf(hdr.Locked));
  if (hdr.Locked and BigFileFlag) <> 0 then
  begin
    Result := Result + stream.Read(UpperLongIntOfSize,sizeof(UpperLongIntOfSize));
    hdr.FullSize := (Int64(UpperLongIntOfSize) Shl 31) + hdr.FullSize;
  end;
end;

function TCompress.WriteHeader(Stream: TStream; var hdr: TCompressHeader): LongInt;  { V8.0 }

  // This function writes the actual header items to the stream
  function WriteHeaderToStream(Stream: TStream; var hdr: TCompressHeader): LongInt;
  var TmpLongInt: LongInt;
      Idx: Integer;
      aByte: Byte;
  begin
    Result := 0;
    for Idx := 0 to 4 do
      Result := Result + Stream.Write(hdr.ComId[Idx], SizeOf(hdr.ComId[Idx]));
    for Idx := 0 to 2 do
      Result := Result + Stream.Write(hdr.ComMethodId[Idx], SizeOf(hdr.ComMethodId[Idx]));
    TmpLongInt := LongInt(hdr.Fullsize and $7FFFFFFF);
    Result := Result + Stream.Write(TmpLongInt, SizeOf(TmpLongInt));
    aByte := Byte(hdr.ArchiveType);
    Result := Result + Stream.Write(aByte, SizeOf(aByte));
    Result := Result + Stream.Write(hdr.Checksum, SizeOf(hdr.Checksum)) +
                       Stream.Write(hdr.Locked, SizeOf(hdr.Locked));
  end;

var
  UpperLongIntOfSize: LongInt;
begin
  if (hdr.Fullsize <= Int64(MaxLongInt)) or     // if Full size is smaller than 2 GB
     (hdr.ArchiveType = caMulti) then           // or this is a MultiFile archive
  begin
    if (hdr.Fullsize > Int64(MaxLongInt)) and
       (hdr.ArchiveType = caMulti) then  // For Multi file archives bigger
      hdr.Fullsize := 0;                 // than 2 GB we ignore hdr.Fullsize
    Result := WriteHeaderToStream(stream, hdr); // this is all we have to do
    Exit;
  end;
  // now we know Full size is bigger than 2 GB, so...
  hdr.Locked := hdr.Locked or BigFileFlag;
  Result := WriteHeaderToStream(stream, hdr);
  UpperLongIntOfSize := LongInt((hdr.FullSize Shr 31));
  Result := Result + stream.Write(UpperLongIntOfSize,sizeof(UpperLongIntOfSize));
end;

function TCompress.SizeOfHeader(hdr: TCompressHeader): Integer;
begin
  if (hdr.Fullsize <= Int64(MaxLongInt)) or (hdr.ArchiveType = caMulti) then
    Result := TCompressHeaderSize
  else
    Result := TCompressHeaderExSize;
end;

{ unlike compressFILE ops, this can leave the result in the destination
  UNCOMPRESSED, WITHOUT A HEADER if that is the optimal result
}
procedure TCompress.Compress(compressedStream, uncompressedStream: Tstream;
                               compressionMethod: TCompressionMethod);
var hdr: TCompressHeader;
    cID : AnsiString; // v9, cID was string
    checksum: Longint;
    CompressedSize: Int64;  { V8.0 upgraded from Longint }
    CompressedStartPosition, Endposition: Int64;  { V8.0 upgraded from Longint }
begin
  FCompressedToPercentage := 100; { e.g. for copy }

  if (compressionMethod = coCustom) and not assigned(FOnCompress) then
    compressionMethod := coNone; { just for safety }

  CompressedStartPosition := CompressedStream.Position;

  if uncompressedStream.size < TCompressHeaderSize then { forget it! }
     CompressionMethod := coNone;

  if (compressionMethod <> coNone) or (Key<>0) then { V4.0 }
  begin
    cID := CompressionCodes[CompressionMethod];
    SetHeader(hdr, cID, caSingle, uncompressedStream.size);
    WriteHeader(CompressedStream, hdr);             { V8.0 - use our own write function }
    if CompressionMethod <> coNone then
      FMustCompressByAtLeast := TCompressHeaderSize { always reset to zero after DoCompress! }
    else { just doing for encryption with no compression V4.0 }
      FMustCompressByAtLeast := 0;
  end;

  try
    try
      CompressedSize := DoCompress(compressedStream, uncompressedStream,
                                      compressionMethod, cID, checksum);
    except on EUnableToCompress do { stopped when reached original size, so save
                                     with no compression and NO header }
      begin
        FMustCompressByAtLeast := 0;
        CompressionMethod := coNone; { ok? }
        unCompressedStream.seek(0,soBeginning);
        if Key<>0 then { V4.0 -- KEEP header if encrypting }
        begin
          compressedStream.Seek(CompressedStartPosition+SizeOfHeader(hdr),soBeginning);
          cID := CompressionCodes[CompressionMethod];
        end else
          compressedStream.Seek(CompressedStartPosition,soBeginning);
        CompressedSize := DoCompress(compressedStream, uncompressedStream,
                                      compressionMethod, cID, checksum); { V4.0, was copyfrom }
      end;
    end; { except }
  finally
    FMustCompressByAtLeast := 0;
  end;
  if (CompressionMethod = coNone) and (Key=0) then exit; { our duty is done, sticks at 100 }

  Inc(CompressedSize,SizeOfHeader(hdr));

  if hdr.FullSize > 0 then
    FCompressedToPercentage := CalcPercentage(CompressedSize,hdr.FullSize);

  { update header -- custom ID, checksum }
  EndPosition := CompressedStream.Position;
  compressedStream.Seek(CompressedStartPosition,soBeginning);
{$IFDEF CLR}
  Move(cID,hdr.ComMethodID);
{$ELSE}
  System.Move(cID[1],hdr.ComMethodID,sizeof(hdr.ComMethodID));
{$ENDIF}
  hdr.checksum := checksum;
  WriteHeader(CompressedStream, hdr);             { V8.0 - use our own write function }
  compressedStream.Seek(EndPosition,soBeginning);
end;

function TCompress.Recognize(const cID: AnsiString) : TCompressionMethod;  // v9, cID was string
var recognized: Boolean;
begin
   for result := low(TCompressionMethod) to high(TCompressionMethod) do
       if cID = CompressionCodes[result] then exit; { incl. custom }

   if assigned(FOnRecognize) then { special case }
   begin
        recognized := False;
        FOnRecognize(cID, recognized);
        if recognized then
        begin
         Result := coCustom;
         exit;
        end;
   end;
   raise EUnrecognizedCompressionMethod.Create('Unrecognized compression method: "'+
                                               cID+'" in header');
end;

procedure TCompress.DoExpand(expandedStream, unexpandedStream: Tstream;
                       compressedsize, fullsize: Int64;  { V8.0 upgraded from Longint }
                       checksum: Longint;
                       compressionMethod: TCompressionMethod; const CompressID: AnsiString;  // v9, CompressID was string
                       FLocked: Longint); { FLocked now Longint V3.5 }
var expandchecksum: Longint;
begin
  if (FLocked and BigFileFlagMask > 0) and (Key=0) then
      raise EInvalidKey.Create('File locked -- valid key required');
  expandchecksum := 0;
  FCompressionTime := GetTickCount;
   if CompressionMethod = coCustom then
   begin
     if assigned(FOnExpand) then
        FOnExpand(expandedStream,unExpandedStream, compressedsize, fullsize, CompressID, expandchecksum);
            { CompressID in case has internal coding, e.g ver }
   end else
     try
       ProcessStreams(expandedStream, unexpandedStream,compressedsize,fullsize,
                      compressionMethod, expandchecksum, cmExpand, FLocked=EncryptFlag);
     except on ELZHBadDecodeTable do { V3.5 }
       if (FLocked and BigFileFlagMask) > 0 then
          raise EInvalidKey.Create('File locked -- valid key required')
       else
          raise; { stick with it... }
     end;

  FCompressionTime := GetTickCount - FCompressionTime;
  if (expandchecksum <> checksum) and
        not ((checksum=0) and (compressionMethod=coNone)) then
    if (FLocked and BigFileFlagMask) > 0 then { TCompress V3.0 or earlier lock check applies if checksum fails }
    begin
      if (expandchecksum+Key) <> checksum then
        raise EInvalidKey.Create('File locked -- valid key required')
    end else { genuine checksum error }
       raise EBadChecksum.Create('Bad checksum during expansion -- data may be invalid');
end;

procedure TCompress.Expand(expandedStream, unExpandedStream: Tstream);
var  hdr: TCompressHeader;
     checksum: Longint;
     compressionMethod: TCompressionMethod;
     StartPosition, csize: Int64;  { V8.0 upgraded from Longint }
     cID: AnsiString;  // v9, cID was string
begin
  compressionMethod:=coNone;
  checksum := 0;
  StartPosition := unExpandedStream.Position;
  csize := unExpandedStream.size-StartPosition;
  if csize >= TCompressHeaderSize then { might have a header  } // v8.3 -- change > to >= so empty encrypted data is *regarded* as encrypted }
  begin
     ReadHeader(unExpandedStream, hdr);           { have a peek; V8.0 - use our own read function }
     if hdr.ComID=CompressedFileHeaderID then     { yup! }
     begin
        if hdr.ArchiveType<>caSingle then { this mode can't handle caMulti! }
           raise EInvalidHeaderArchiveType.Create('Expand can''t handle caMulti archive type in compressed header');
        SetLength(cID,sizeof(hdr.ComMethodID));
{$IFDEF CLR}
        Move(hdr.ComMethodID,cID);
{$ELSE}
        System.Move(hdr.ComMethodID,cID[1],sizeof(hdr.ComMethodID));
{$ENDIF}
        CompressionMethod := Recognize(cID); { raises exception if can't recognize }
        checksum := hdr.checksum;
        Dec(csize, SizeOfHeader(hdr));
        if expandedStream is TMemoryStream then { get the size in one go }
           TMemoryStream(expandedStream).setSize(TMemoryStream(expandedStream).size
                           +hdr.FullSize); { V4.0 added existing size in in case non-0 }
     end else { nope -- not compressed, no header, copy the LOT}
     begin
        unexpandedStream.Seek(StartPosition,soBeginning); { rewind for full copy }
        expandedStream.copyFrom(unExpandedStream,csize);
        exit;
     end;
  end else
     hdr.Locked := 0; { ensure not set }
  { note: hdr.Fullsize undefined if coNone }
  DoExpand(expandedStream, unexpandedStream,
           csize,hdr.Fullsize,checksum,compressionMethod, cID, hdr.Locked);
end;


{$IFDEF CLR}
function TCompress.GetFileHeader(Stream: TStream;var Fhdr: TCompressedFileHeader; var Filename: String): Boolean;
var BytesRead: LongInt;
    tmpStrBuff: array of Byte;
begin
  result := False;
  BytesRead := ReadFileHeader(stream, Fhdr);             { V8.0 - use our own read function }
  if BytesRead = SizeOfFileHeader(Fhdr) then             { got one }
  begin
   SetLength(Filename,Fhdr.FilenameLength);   // v9 - please note, FileNameLength is in BYTES
   setLength(tmpStrBuff, Fhdr.FilenameLength);
   BytesRead := stream.read(tmpStrBuff, Fhdr.FilenameLength);
   Filename := StringOf(tmpStrBuff);
   if BytesRead = Fhdr.FilenameLength then
      Result := True;
  end;
end;
{$ELSE}
function TCompress.GetFileHeader(Stream: TStream;var Fhdr: TCompressedFileHeader; var Filename: String): Boolean;
var BytesRead: LongInt;
    tmpStr: AnsiString; // v9
begin
  result := False;
  BytesRead := ReadFileHeader(stream, Fhdr);             { V8.0 - use our own read function }
  if BytesRead = SizeOfFileHeader(Fhdr) then             { got one }
  begin
   SetLength(tmpStr,Fhdr.FilenameLength);   // v9 - note, FileNameLength is in BYTES
   if stream.read(tmpStr[1],Fhdr.FilenameLength)=Fhdr.FileNameLength then // v9 was Filename[1]
   begin
      Result := True;
      Filename := tmpStr;   // v9 - effectively, we are supporting only AnsiString filenames, for backwards compatibility
   end;
  end;
end;
{$ENDIF}


{$IFDEF CLR}
procedure TCompress.PutFileHeader(Stream: TStream;var Fhdr: TCompressedFileHeader; const Filename: String);
begin
  Fhdr.FilenameLength := Length(Filename));
  WriteFileHeader(stream, Fhdr);                         { V8.0 - use our own write function }
  stream.writebuffer(BytesOf(Filename), Length(Filename)));
end;
{$ELSE}
procedure TCompress.PutFileHeader(Stream: TStream;var Fhdr: TCompressedFileHeader; const Filename: String);
var tmpStr: AnsiString; // v9
begin
  tmpStr := Filename; // v9 - we're storing AnsiString filename only, for backwards compatibility
  Fhdr.FilenameLength := Length(tmpStr); // v9, was Length(Filename)
  WriteFileHeader(stream, Fhdr);                         { V8.0 - use our own write function }
  stream.writebuffer(tmpStr[1],Length(tmpStr)); // v9 was Filename[1], Length(Filename)
end;
{$ENDIF}

function TCompress.ReadFileHeader(Stream: TStream; var Fhdr: TCompressedFileHeader): LongInt;   { V8.0 }
var UpperLongIntOfSize: LongInt;
    TmpLongInt: LongInt;
    aByte: Byte;
begin
  Result := Stream.Read(Fhdr.FileNameLength, SizeOf(Fhdr.FileNameLength)) +
            Stream.Read(Fhdr.Datetime, SizeOf(Fhdr.Datetime)) +
            Stream.Read(Fhdr.Attributes, SizeOf(Fhdr.Attributes));
  Result := Result + Stream.Read(TmpLongInt, SizeOf(TmpLongInt));
  Fhdr.Fullsize := Int64(TmpLongInt);
  Result := Result + Stream.Read(TmpLongInt, SizeOf(TmpLongInt));
  Fhdr.CompressedSize := Int64(TmpLongInt);
  Result := Result + Stream.Read(aByte, SizeOf(aByte));
  Fhdr.CompressedMode := TCompressionMethod(aByte);
  Result := Result + Stream.Read(Fhdr.Checksum, SizeOf(Fhdr.Checksum)) +
                     Stream.Read(Fhdr.Locked, SizeOf(Fhdr.Locked));
  if (Fhdr.Locked and BigFileFlag) <> 0 then
  begin
    Result := Result + stream.Read(UpperLongIntOfSize,sizeof(UpperLongIntOfSize));
    Fhdr.FullSize := (Int64(UpperLongIntOfSize) Shl 31) + Fhdr.FullSize;
    Result := Result + stream.Read(UpperLongIntOfSize,sizeof(UpperLongIntOfSize));
    Fhdr.CompressedSize := (Int64(UpperLongIntOfSize) Shl 31) + Fhdr.CompressedSize;
  end;
end;

function TCompress.WriteFileHeader(Stream: TStream; var Fhdr: TCompressedFileHeader): LongInt;  { V8.0 }

  // This function writes the actual file header items to the stream
  function WriteFileHeaderToStream(Stream: TStream; var Fhdr: TCompressedFileHeader): LongInt;   { V8.0 }
  var TmpLongInt: LongInt;
      aByte: Byte;
  begin
    Result := Stream.Write(Fhdr.FileNameLength, SizeOf(Fhdr.FileNameLength)) +
              Stream.Write(Fhdr.Datetime, SizeOf(Fhdr.Datetime)) +
              Stream.Write(Fhdr.Attributes, SizeOf(Fhdr.Attributes));
    TmpLongInt := LongInt(Fhdr.FullSize and $7FFFFFFF);
    Result := Result + Stream.Write(TmpLongInt, SizeOf(TmpLongInt));
    TmpLongInt := LongInt(Fhdr.CompressedSize and $7FFFFFFF);
    Result := Result + Stream.Write(TmpLongInt, SizeOf(TmpLongInt));
    aByte := Byte(Fhdr.CompressedMode);
    Result := Result + Stream.Write(aByte, SizeOf(aByte));
    Result := Result + Stream.Write(Fhdr.Checksum, SizeOf(Fhdr.Checksum)) +
                       Stream.Write(Fhdr.Locked, SizeOf(Fhdr.Locked));
  end;

var
  UpperLongIntOfSize: LongInt;
begin
  if Fhdr.Fullsize <= Int64(MaxLongInt) then           // if Full size is smaller than 2 GB
  begin
    Result := WriteFileHeaderToStream(Stream, Fhdr);   // this is all we have to do
    Exit;
  end;
  // now we know Full size is bigger than 2 GB, so...
  Fhdr.Locked := Fhdr.Locked or BigFileFlag;
  Result := WriteFileHeaderToStream(Stream, Fhdr);
  UpperLongIntOfSize := LongInt((Fhdr.FullSize Shr 31));
  Result := Result + stream.Write(UpperLongIntOfSize,sizeof(UpperLongIntOfSize));
  UpperLongIntOfSize := LongInt((Fhdr.CompressedSize Shr 31));
  Result := Result + stream.Write(UpperLongIntOfSize,sizeof(UpperLongIntOfSize));
end;

function TCompress.SizeOfFileHeader(Fhdr: TCompressedFileHeader): Integer;
begin
  if Fhdr.Fullsize <= Int64(MaxLongInt) then
    Result := TCompFileHeaderSize
  else
    Result := TCompFileHeaderExSize;
end;

procedure TCompress.Move(source: string; var dest: array of AnsiChar);
var idx: Integer;
begin
  for idx := 0 to Length(source)-1 do
    dest[idx] := AnsiChar(source[idx+1]);
end;

procedure TCompress.Move(source: array of AnsiChar; var dest: string);
var idx: Integer;
begin
  dest := '';
  for idx := 0 to Length(source)-1 do
    dest := dest + AnsiChar(source[idx]);
end;

procedure TCompress.ScanCompressedStream(compressedstream: TStream;
                                                   var Finfo: TStringList);
var hdr: TCompressHeader;
    Fhdr: TCompressedFileHeader;
    Filename: String;
    cfileinfo : TCompressedFileInfo;
    cID: AnsiString;  // v9, cID was string
begin
  if (compressedStream.size-compressedStream.position) >= TCompressHeaderSize+TCompFileHeaderSize then { got something in it }
  begin
    CheckHeader(compressedStream,hdr,cID);

    while GetFileHeader(compressedStream,Fhdr,Filename) do { got one }
    begin
     cfileinfo := TCompressedFileInfo.Create;
     cfileinfo.Datetime := Fhdr.Datetime;
     cfileInfo.Attributes := FHdr.Attributes;
     cfileinfo.Fullsize := Fhdr.Fullsize;
     cfileinfo.Compressedsize := Fhdr.CompressedSize;
     cfileinfo.Position := compressedStream.position-SizeOfFileHeader(Fhdr)-Fhdr.FilenameLength;
     cfileinfo.CompressedMode := Fhdr.CompressedMode;
     cfileinfo.Checksum := Fhdr.Checksum;
     cfileInfo.Locked := (FHdr.Locked and BigFileFlagMask) >0;
     FInfo.addobject(filename,cfileinfo);
     compressedStream.seek(Fhdr.Compressedsize,soCurrent); { move on... }
    end;
  end;
end;

procedure TCompress.ScanCompressedFile(const arcfile: String; var Finfo: TStringList);
var cfile: TFilestream;
begin
   cfile :=TFilestream.create(arcfile,fmOpenRead {$IFNDEF Linux}or fmShareExclusive{$ENDIF});
   try
     ScanCompressedStream(cfile,Finfo);
   finally
     cfile.free;
   end;
end;

{ Assumes both exist on same path. Renames new to old AND deletes old... }
procedure replaceFile(oldfile, newfile: String);
begin
  SysUtils.DeleteFile(oldfile);
  renameFile(newfile,oldfile);
end;

function FixStoredFilename(s: String): String;
begin
  if (Length(s) > 2) and (s[2]=':') then { has a drive specifier -- get rid of it }
       s := Copy(s,3,Length(s));
  result := s; {AnsiLowerCase(s); No longer changing case, V3.0 }
end;


{ added V3.0 to avoid changing filename case }
function TCompress.FileInList(whichFiles: TStringList;filename: string): Boolean;
var count: Integer;
begin
  result := True;
  for count := 0 to pred(whichFiles.count) do
{$IFDEF Linux}
  if (FIgnoreCase and (CompareText(Filename,whichFiles[count])=0)) or
     (Filename=whichFiles[count]) then
{$ELSE}
    if CompareText(Filename,whichFiles[count])=0 then
{$ENDIF}
      exit;
  result := False; { not there }
end;

{ The cute thing about this is that it does NOT insist on
  destfile pre-existing, so can implement a DELETE with it too! }
procedure TCompress.AppendFilesExcept(const destfile,fromfile: String;
                                                      notfiles: TStringlist);
var dest, source: TFileStream;
  dhdr, shdr: TCompressHeader;
  FHdr: TCompressedFileHeader;
  Filename: String;
begin
  source:=TFilestream.create(fromfile,fmOpenRead {$IFNDEF Linux}or fmShareExclusive{$ENDIF});

  if FileExists(destfile) then
    dest:=TFilestream.create(destfile,fmOpenReadWrite {$IFNDEF Linux}or fmShareExclusive{$ENDIF})
  else  { create from scratch -- e.g. in delete case }
    dest:=TFilestream.create(destfile,fmCreate);

  try
    if source.size > TCompressHeaderSize + TCompFileHeaderSize then { ok to proceed }
    begin
        ReadHeader(source, shdr);               { V8.0 - use our own read function }
        if dest.size > TCompressHeaderSize then
        begin
          ReadHeader(dest, dhdr);               { V8.0 - use our own read function }
          dest.seek(0,soEnd); { end of file }
        end else { no header -- make one like source }
        begin
          dhdr := shdr;
          dhdr.Fullsize:= 0; { for now... }
          WriteHeader(dest, dhdr);              { V8.0 - use our own write function }
        end;

        while GetFileHeader(source,Fhdr,Filename) do { got one }
        begin
           if not FileInList(notfiles,Filename) then { 3.02 keep? ok, copy to dest }
           begin
             PutFileHeader(dest,Fhdr,Filename);
             if fhdr.CompressedSize> 0 then dest.Copyfrom(source,fhdr.CompressedSize);
             Inc(dhdr.Fullsize,fhdr.FullSize);
           end else
             source.seek(Fhdr.Compressedsize,soCurrent);
        end;
        dest.seek(0,soBeginning);
        WriteHeader(dest, dhdr);                { V8.0 - use our own write function }
    end;
  finally
   dest.free;
   source.free;
  end;
end;


procedure TCompress.DeleteFiles(const arcFile: String; whichfiles: TStringList);
var newfilename, storedwhichfile, whichfile: String;
     count,here: Integer;
begin
  count := 0;
  while count < whichfiles.count do
  begin
     whichfile := whichfiles[count];
     storedwhichfile := whichfile;
     if assigned(FOnCheckFile) then
     begin
        FOnCheckFile(storedwhichfile,cmDelete); { changes as required... }
        if (storedwhichFile=CompressSkipFlag) or (storedwhichfile=CompressNoMoreFlag) then
        begin
          whichfiles.delete(count); { modify the list for reference by caller... }
          if (storedwhichfile=CompressNoMoreFlag) then
          begin
             here := count;
             for count:=whichfiles.count-1 downto here+1 do { any more? }
              whichfiles.delete(count);
             break; { we've emptied the remainder of the list... }
          end;
          continue;                { skip increment }
        end;
     end;
     inc(count);
  end; { for }
  { We've now verified which ones to delete... go ahead and do it... }
  if whichfiles.count > 0 then
  begin
    newfilename := ChangeFileExt(arcfile,TempExtension); { too bad if it exists... }
    AppendFilesExcept(newfilename,arcfile,whichfiles); { see? }
    ReplaceFile(arcfile,newfilename); { just like that ! }
  end;
end;

procedure TCompress.CompressFiles(const arcFile: String; whichfiles: TStringList;
                                            CompressionMode: TCompressionMethod);
var dest: TFileStream;
     update: Boolean;
     newfilename: String;
begin
  update:=False;
  newfilename := arcfile;
  if FileExists(arcfile) then { we are updating an existing archive... I assume }
  begin
    update:=True;
    newfilename := ChangeFileExt(arcfile,TempExtension); { overwritten if it exists... }
  end;
 { special check for what should be obvious reasons (and fmShareExclusive seems not to lock OURSELVES out!) }
  if FileInList(whichfiles,newfilename) then { V3.02 FileInList}
{$IFDEF Linux}
     if not FIgnoreCase then
       whichfiles.delete(whichfiles.indexof(newfilename))
     else
{$ENDIF}
     whichfiles.delete(whichfiles.indexof(AnsiLowerCase(newfilename)));
      { of course, they could CHEAT & force it in FonCheck file but... }

  dest:=TFilestream.create(newfilename,fmCreate {$IFNDEF Linux}or fmShareExclusive{$ENDIF});
  try
    CompressFilesToStream(dest,whichfiles,CompressionMode);
  except on EWriteError do  { ran out of disk space? or protected? }
    begin
      dest.free;
      SysUtils.DeleteFile(newfilename);
      raise EInsufficientDiskSpace.Create('Disk protected or full during compression');
    end;
  end;

  dest.free;
  if update then begin { lots to do here... }
     try
       AppendFilesExcept(newfilename,arcfile,whichfiles);
     except on EWriteError do  { ran out of disk space? }
         raise EInsufficientDiskSpace.Create('Disk full or write error during compression');
     end;
     ReplaceFile(arcfile,newfilename);
  end;
end;

{ Note: This function DOES NOT update/append to an existing stream --
  that would have to be managed by the calling procedure (see CompressFiles).
In other words, it creates a NEW archive at the current position in dest   }
function TCompress.CompressFilesToStream(dest: TStream; var whichfiles: TStringList;
                                      CompressionMode: TCompressionMethod): Int64;  { V8.0 upgraded from Longint }
var hdr: TCompressHeader;
    fromfile, storedfromfile: String;
    cID: AnsiString;  // v9, cID was string
    count : Integer;
    source: TFileStream;
    Fhdr: TCompressedFileHeader;
    Filename: String;
    saveDestStart: Int64; { V4.0, thanks to Bill Sherman } { V8.0 upgraded from Longint }
begin
  saveDestStart := dest.position; { V4.0 }
  source := nil; { V3.0 to silence warning }
  cID := CompressionCodes[CompressionMode];
  Result := 0; { non-zero means dest is a FILE requiring truncation after dest is closed... }
  SetHeader(hdr, cID, caMulti, Longint(0));
  WriteHeader(dest, hdr);             { V8.0 - use our own write function }
  count := 0;
  while count < whichFiles.count do
  begin
     fromfile := whichfiles[count];
     storedfromfile := fromfile;
     if TargetPath<>'' then { trim off target path if set?  V2.5 }
       if CompareText(Copy(storedFromFile,1,Length(targetPath)),targetPath)=0 then { cut it }
             storedFromFile :=Copy(storedFromFile,Length(targetPath)+1,9999);
     if assigned(FOnCheckFile) then
     begin
        FOnCheckFile(storedFromFile,cmCompress); { changes as required... }
        if (storedFromFile=CompressSkipFlag) or (storedFromfile=CompressNoMoreFlag) then
        begin
          whichfiles.delete(count); { modify the list for reference by caller... }
          if (storedFromfile=CompressNoMoreFlag) then
             break; { note that earlier files ARE processed }
          continue;                { skip it }
        end;
     end;
     storedfromFile := FixStoredFilename(storedfromFile); { no drive, all lower }
     try
       source:=TFilestream.create(fromfile,fmOpenRead or fmShareDenyNone); {V3.0 (was fmShareExclusive) }
     except on EFOpenError do
      begin
        if ExceptionOnFileError then
           raise;
        whichfiles.delete(count); { modify the list for reference by caller... }
        continue;                { skip it }
      end;
     end;
     whichfiles[count] := storedfromfile; { save amendments for reference by caller }
     Filename := storedfromfile;

     with FHdr do
     begin
{$IFDEF CLR}
       DateTime := FileAge(fromfile);
{$ELSE}
       DateTime := FileGetDate(source.Handle);
{$ENDIF}
       Attributes := fileGetattr(fromfile);
     end;

     try { V3.0 rewrite to use new routine... }
       inc(hdr.Fullsize,source.size);
       CompressStreamToArchiveStream(dest,source,
            Filename,FHdr,cID,CompressionMode);
     finally
       source.free;
     end;

     inc(Count);
  end; { for }

{$IFDEF CLR}
  Move(cID,hdr.ComMethodID);                                    { in case CUSTOM changed it... }
{$ELSE}
  System.Move(cID[1],hdr.ComMethodID,sizeof(hdr.ComMethodID));  { in case CUSTOM changed it... }
{$ENDIF}
  dest.seek(saveDestStart,soBeginning); { saveDestStart V4.0, was 0 }
  WriteHeader(dest, hdr);               { updates Fullsize as well; V8.0 - use our own write function }
end;

{ New in V3.0 --
 Called to write an arbitrary stream to the current archive position.
 Assumes:
a) FHdr.DateTime, FHdr.attributes and Filename are all valid
b) both source and dest are at the correct starting place
c) cID is at least 3 bytes long, potentially holding a custom compression method ID
   -- *should* be written to archive header's ComMethodID field when all done
e) Archive header's Fullsize value will be incremented with the source.size-source.position value
NB: will save WITHOUT compression if saving WITH makes it larger

SETS following Fhdr fields:
 CompressedMode, CompressedSize, FullSize, FileNameLength, Locked
}

procedure TCompress.CompressStreamToArchiveStream(compressedStream, unCompressedStream: TStream;
            const Filename: string; FHdr: TCompressedFileHeader;
            var cID: AnsiString; CompressionMode: TCompressionMethod); // v9 cID was string
var  fhdrpos, savepos, thisfilein, thisfileout: Int64;  { V8.0 upgraded from Longint }
begin
     savepos := uncompressedStream.position;
     thisfilein := uncompressedStream.size-savepos;
     fhdrpos := compressedStream.position;
     fhdr.Fullsize := thisfilein;     { V8.0 - moved here from [with FHdr do] bellow }
     fhdr.Locked := 0;                { V8.0 - moved here from [with FHdr do] bellow }

     PutFileHeader(compressedStream,Fhdr,Filename);

     try
       thisfileout := DoCompress(compressedStream,uncompressedStream,compressionMode,cID,fhdr.checksum);
     except on EUnableToCompress do { stopped when reached original size, so save
                                      with no compression and NO header }
       begin
         CompressionMode := coNone; { flag }
         compressedStream.Seek(fhdrpos+SizeOfFileHeader(Fhdr)+Fhdr.FilenameLength,soBeginning); { rewrite from compressed area }
         uncompressedStream.seek(savepos,soBeginning); { rewind! }
         thisfileout := DoCompress(compressedStream, uncompressedStream,
                                   compressionMode,cID, fhdr.checksum); { V4.0, was copyfrom }
       end;
     end; { except }

     savepos := compressedStream.position;
     with FHdr do
     begin
       CompressedSize := thisfileout;
       CompressedMode := CompressionMode; { the default for this Cfile }
       if Key > 0 then
         Locked := Locked or EncryptFlag; { V3.5 }
     end;       { "Locked or" added in V8.0 }

     compressedStream.seek(fhdrpos,soBeginning);
     WriteFileHeader(compressedStream, fhdr);            { V8.0 - use our own write function }
     compressedStream.seek(savepos,soBeginning);
     if thisfilein > 0 then
       FCompressedToPercentage := CalcPercentage(thisfileout,thisfilein)
     else
      FCompressedToPercentage := 100;
end;


{ Compresses the remaining contents of uncompressedStream using dummy name
  in Filename onto the end of the current archive (creates if necessary).
  Does NOT check if Filename is already in the archive.
}
procedure TCompress.CompressStreamToArchive(arcFile: String; uncompressedStream: TStream; const fileName: string;
                           CompressionMode: TCompressionMethod); { V3.0 }
var dest: TFileStream;
    cID: AnsiString; // v9 AnsiString was string
    hdr: TCompressHeader;
    fHdr: TCompressedFileHeader;
begin
  cID := CompressionCodes[CompressionMode];
  if FileExists(arcFile) then
  begin
    dest:=TFilestream.create(arcFile,fmOpenReadWrite {$IFNDEF Linux}or fmShareExclusive{$ENDIF});
    try
      CheckHeader(dest,hdr,cID);
      dest.seek(0,soEnd); { always appends, no checks on pre-existence }
    except on Exception do begin dest.free; raise end
    end;
  end else { new archive }
  begin
    dest:=TFilestream.create(arcFile,fmCreate {$IFNDEF Linux}or fmShareExclusive{$ENDIF});
    SetHeader(hdr, cID, caMulti, Longint(0));
    // ensure we have the correct header in the archive
    WriteHeader(dest, hdr);               { V8.0 - use our own write function }
  end;

  try
    try
      fHdr.Datetime := 0;
      fHdr.attributes :=0; { all else is set & written in the next call... }
      inc(hdr.Fullsize,uncompressedStream.size-uncompressedStream.position);
      CompressStreamToArchiveStream(dest,uncompressedStream,fileName,fHdr,
                                    cID,CompressionMode);
{$IFDEF CLR}
      Move(cID,hdr.ComMethodID);                                    { in case CUSTOM changed it... }
{$ELSE}
      System.Move(cID[1],hdr.ComMethodID,sizeof(hdr.ComMethodID));  { in case CUSTOM changed it... }
{$ENDIF}
      dest.seek(0,soBeginning);
      // updates Fullsize as well
      WriteHeader(dest, hdr);               { V8.0 - use our own write function }
    except on EWriteError do  { ran out of disk space? or protected? }
      raise EInsufficientDiskSpace.Create('Disk protected or full during compression');
    end; { except }
  finally
    dest.free;
  end;
end;

{ Finds FIRST occurrence of Filename in the archive and expand into uncompressedStream. }
procedure TCompress.ExpandStreamFromArchive(arcFile: String; uncompressedStream: TStream; fileName: string); { V3.0 }
var from: TFileStream;
    thisfile: string;
    cID: AnsiString; // v9 was string
    hdr: TCompressHeader;
    fHdr: TCompressedFileHeader;
    ExpandedSize: Int64; { just in case stream continues past archive end }
                  { V8.0 upgraded from Longint }
begin
  from:=TFilestream.create(arcfile,fmOpenRead {$IFNDEF Linux}or fmShareExclusive{$ENDIF});
  try
    CheckHeader(from,hdr,cID);
    ExpandedSize :=0;
    while (GetFileHeader(from,Fhdr,thisfile))
      and ((ExpandedSize < hdr.FullSize) or
           ((hdr.ArchiveType = caMulti) and (hdr.Fullsize = 0))) do
    begin
     if CompareText(thisfile,Filename)=0  then { V3.03 -- not case sensitive, stop when found }
     begin
       DoExpand(unCompressedStream,from,fhdr.CompressedSize,fhdr.Fullsize,fhdr.checksum,
                                fhdr.CompressedMode,cID, Fhdr.Locked)
     end else
        from.seek(Fhdr.Compressedsize,soCurrent); { move on... }
     Inc(ExpandedSize,fhdr.Fullsize);
    end; { while }
  finally
    from.free;
  end;
end;


procedure TCompress.ExpandFiles(const arcfile:String;whichfiles: TStringList);
var from: TFilestream;
begin
  from:=TFilestream.create(arcfile,fmOpenRead {$IFNDEF Linux}or fmShareExclusive{$ENDIF});
  try
     ExpandFilesFromStream(from,whichfiles);
  finally
     from.free;
  end;
end;

{ added v4.0 to overcome Delphi's diskfree 2GB limit.
  From The Delphi Magazine Tips 'n Tricks dept }
{$IFDEF MSWINDOWS}

{$IFDEF UNICODE} // v9, D2009 or later
function TCompress.DiskFree(Drive:byte): Int64;
begin
  Result := Sysutils.DiskFree(Drive);
end;
{$ELSE} // Windows versions earlier than D2009+

function TCompress.DiskFree(Drive:byte): Int64;
var
  RootPath: array[0..4] of AnsiChar;
  RootPtr : Pchar;
  SectorsPerCluster,bytesPerSector,
  FreeClusters, TotalClusters: DWORD; { V4.0, was Integer }
  SPC: Single;
begin
  RootPtr := Nil;
  if Drive > 0  then
  begin
    StrCopy(RootPath, 'A:\');
    RootPath[0] := AnsiChar(Drive +$40);
    RootPtr := RootPath;
  end;
  If GetDiskFreeSpace(RootPtr,SectorsPerCluster,BytesPerSector,
    FreeClusters,TotalClusters) then
  begin
    SPC := SectorsPerCluster;
    Result := Trunc(SPC) * Int64(BytesPerSector) * Int64(FreeClusters);
  end else
    Result := -1;
end;
{$ENDIF} // ELSE UNICODE
{$ENDIF} // MSWINDOWS

{$IFDEF LINUX}
function TCompress.DiskFree(Drive:byte): Int64;
var
  DriveInfo : TStatVFS;
begin
  StatVFS('/', DriveInfo);
  Result := Int64(DriveInfo.f_bsize) * Int64(Driveinfo.f_bfree);
end;
{$ENDIF}

{$IFDEF CLR}
function TCompress.DiskFree(Drive:byte): Int64;
begin
  Result := Sysutils.DiskFree(Drive);
end;
{$ENDIF}

{$IFDEF CLR}
function TCompress.CheckDiskSpace(filename: string): Int64;
var diskid: Integer;
    freespace: Int64; { was Longint 2.5b }
begin
  result := ConstMaxINT64; { automatic pass }
  if not CheckSpaceBeforeExpand then
     exit;
  diskid := 0;
  filename := uppercase(filename);
  if (length(filename)>1) and (filename[2]=':') then { has drive }
     diskid := ord(filename[1])-ord('@'); { 1 for A, 2 for B etc. }
  freespace := diskfree(diskid); { so far so good }
  if freespace < ConstMaxINT64 then { was freespace > 0, 2.5b }
    result := freespace; { otherwise result stays at MaxInt64, which is fine }
end;
{$ELSE}
{ this function revised in v4.0 to use new diskfree }
function TCompress.CheckDiskSpace(filename: string): Int64;  { V8.0 upgraded from Longint }
var  diskid: Integer;
     f : file of Byte;
     fsize: Cardinal;
     freespace: single; { was Longint 4.0 }
begin
  result := ConstMaxINT64; { automatic pass }
            { V8.0 updated from MaxLongInt}
  if not CheckSpaceBeforeExpand then
     exit;
  diskid:=0;
{$IFNDEF Linux}
  filename:=uppercase(filename);
  if (length(filename)>1) and (filename[2]=':') then { has drive }
     diskid:=ord(filename[1])-ord('@'); { 1 for A, 2 for B etc. }
{$ENDIF}
  freespace := diskfree(diskid); { so far so good }
  if freespace < ConstMaxINT64 then { was freespace > 0, 4.0 }
                 { V8.0 updated from MaxLongInt}
    result := Trunc(freespace); { otherwise result stays at MaxInt64, which is fine }

  if (result < ConstMaxINT64) and FileExists(filename) then
               { V8.0 updated from MaxLongInt}
  begin
{$IFDEF Linux}
   fsize := FileSizeByName(filename);
{$ELSE}
   Assignfile(f,filename); { open it }
   Filemode := 0;
   Reset(f);
   fsize := Filesize(f);
   Closefile(f); { and that's it... }
{$ENDIF}
   if Freespace+fsize < ConstMaxINT64 then   { use Freespace v4.0 }
                        { V8.0 updated from MaxLongInt}
     result := result+fsize
   else
     result := ConstMaxINT64; { V8.0 updated from MaxLongInt}
  end;
end;
{$ENDIF}

{ Validates that stream (forced to the start) is:
a) Big enough to hold a TCompress header (else no action -- leaves at start and
   initializes hdr to caMulti/current cID)
b) Contains valid CompressedFileHeaderID
c) Is of type caMulti
d) has a recognisable ComMethodID, e.g. 'LZH'
   -- raises exceptions if not. If ok, hdr and cID are set and the stream
   position is at the end of the header
}
procedure TCompress.CheckHeader(compressedFile: TStream;var hdr: TCompressHeader; var cID: AnsiString); // v9 cID was string
begin
   SetLength(cID,sizeof(hdr.ComMethodID));
{   compressedFile.seek(0,soBeginning);  disabled v3.5 }
   if (compressedFile.size-compressedFile.position) >= SizeOfHeader(hdr) then
   begin
     ReadHeader(compressedFile, hdr);                     { V8.0 - use our own read function }
     if hdr.ComID<>CompressedFileHeaderID then { oops }
       raise EInvalidHeader.Create('Can''t recognize compressed file header');
     if hdr.ArchiveType<>caMulti then { wrong mode! }
       raise EInvalidHeaderArchiveType.Create('Archive type must be caMulti');
{$IFDEF CLR}
    move(hdr.ComMethodId, cID);
{$ELSE}
    System.Move(hdr.ComMethodID,cID[1],sizeof(hdr.ComMethodID));
{$ENDIF}
    Recognize(cID); { raises exception if can't recognize }
   end else
     SetHeader(hdr, cID, caMulti, Longint(0)); { cID *may* be garbage if not initialised before call }
end;

procedure TCompress.ExpandFilesFromStream(compressedstream:TStream;whichfiles: TStringList);
var hdr: TCompressHeader;
    FHdr: TCompressedFileHeader;
    Filename: String;
    dest: TFilestream;
    whichfile: String;
    ExpandedList: TStringlist;
    expandedSize: Int64; { V2.5 so we can stop even if resource size > actual archive size }
                  { V8.0 upgraded from Longint }
    cID: AnsiString; // v9, cID was string
begin
  dest := nil; { V3.0 to silence warning }
  ExpandedList:=TStringlist.Create;
  try
    CheckHeader(compressedStream,hdr,cID);

    ExpandedSize := 0;
    while (GetFileHeader(compressedStream,Fhdr,Filename))
      and ((ExpandedSize < hdr.FullSize) or                         { V2.5 added ExpandedSize test }
           ((hdr.ArchiveType = caMulti) and (hdr.Fullsize = 0))) do { V8.0 test ignored for MultiFile archives with Fullsize = 0}
    begin
      if (whichfiles<>nil) and (not FileInList(whichFiles,filename)) then  { FileInList V3.0 }
         compressedStream.seek(Fhdr.Compressedsize,soCurrent) { move on... }
      else
      begin
        whichfile := Filename;
        if Copy(whichfile,1,1)<>slash then { not an absolute path?  V2.5 }
           whichfile:=TargetPath+whichfile;  { TargetPath prepend V2.5 }
        if assigned(FonCheckFile) then
        begin
          FOnCheckFile(whichfile,cmExpand); { changes as required... }
          if (whichFile=CompressSkipFlag) then
          begin
            compressedStream.seek(Fhdr.Compressedsize,soCurrent); { move on... }
            continue
          end;
          if whichfile=CompressNoMoreFlag then
              break;
        end;
        { check we have space available -- roughly }

        if CheckDiskSpace(whichfile) < FHdr.fullsize then
           raise EInsufficientDiskSpace.Create('Not enough disk space to expand '+whichfile);

        if MakeDirectories then
           ForceDirectories(ExtractFilePath(Whichfile));

        { here we get to create the file and expand into it... }
        try
          dest:=TFileStream.Create(whichfile,fmCreate {$IFNDEF Linux}or fmShareExclusive{$ENDIF})
        except on EFOpenError do
           begin
              if ExceptionOnFileError then
                 raise;
              compressedStream.seek(Fhdr.Compressedsize,soCurrent); { move on... }
              continue
           end;
        end;
        { Note that we SUPPORT varying modes here, i.e. we use each file's mode,
          BUT the cID could conceivably overrule it if inspected in OnExpand... (we don't) }
        try
          try  { EInvalidKey handling V2.5 }
            DoExpand(dest,compressedStream,
                     fhdr.CompressedSize, fhdr.Fullsize,fhdr.checksum,
                     fhdr.CompressedMode, cID, Fhdr.Locked);
          except on EInvalidKey do
            begin
              dest.free; { close }
              dest := nil;
              SysUtils.DeleteFile(whichfile);
              raise; { raise again so the world knows... }
            end;
          end;
{$IFNDEF Linux} { see below for Linux Setting }
  {$IFDEF CLR}
          if Fhdr.DateTime <> 0 then
            FileSetDate(whichfile,Fhdr.DateTime);
  {$ELSE}
          if Fhdr.DateTime <> 0 then
            FileSetDate(dest.handle,Fhdr.DateTime);
  {$ENDIF}
{$ENDIF}
          ExpandedList.add(whichfile);
        finally
           if dest<>nil then { V2.5 -- add test as possible EInvalidKey }
           begin
            dest.free;
{$IFDEF Linux}
            if Fhdr.DateTime <> 0 then
              FileSetDate(whichfile,Fhdr.DateTime);
{$ELSE}  { V6.01 can't set attributes in Kylix at this point }
            FileSetAttr(whichfile,FHdr.Attributes);
{$ENDIF}
           end;
        end;
      end; { if }
      Inc(ExpandedSize,fhdr.Fullsize); { moved from above FilesetDate V3.0 }
    end; { while }
    if whichFiles<>nil then  { nil test added V2.5 }
    begin
      whichfiles.clear;
      whichfiles.assign(ExpandedList);
    end;
 finally
  ExpandedList.free;
 end;
end;


function TCompress.GetCompressedPercentage: Integer;
begin
   Result := 100-FCompressedToPercentage; { display per PKZIP's approach, i.e. 99% is LOTS }
end;

{ from here on down are low-level routines with ThreadSafe protection of globals }

function TCompThreadSafety.Encrypt(ch: AnsiChar): AnsiChar; { V3.5 }
begin
  result := AnsiChar(ord(ch) xor EncryptCodes[EncryptCounter]);
  inc(EncryptCounter);
  EncryptCounter := EncryptCounter mod 4;
end;

function TCompThreadSafety.GetChar: AnsiChar;
var BytesRead: Int64;  { V8.0: upgraded from Longint }
    Progress: Longint; { V8.0 keep it as Longint for backward compatibility with existing projects }
begin
  if (inBuffPos = 0) or (inBuffPos = lChunk) then
  begin
    Result := #0;
    inBuffPos := 0;
    InputEOF := True; { precautionary }
    if ReadSize = 0 then
      Exit; { no more boss }
    if lChunk > ReadSize then
      lChunk := ReadSize;
{$IFDEF CLR}
    BytesRead := source.Read(inBuffer, lChunk); { read chunk }
{$ELSE}
    BytesRead := source.Read(inBuffer[0], lChunk); { read chunk }
{$ENDIF}
    ReadSize := ReadSize - BytesRead;
    if BytesRead = 0 then { EOF }
      Exit;

    if assigned(GlobalFOnShowProgress) then
    begin
      BytesIn := BytesIn + BytesRead; { read so far -- readsize is "yet to read" }
      Progress := CalcPercentage(BytesIn,BytesIn+Readsize); { V3.01 }
      GlobalFOnShowProgress(Progress);
      if Progress = -1 then { user cancel }
        exit; { InputEOF is still true at present, so we fake an early EOF, is all }
    end;
    Sleep(0);  { V2.5 replaces Application.ProcessMessages so Compress can be
                 used without Forms dependency }
    InputEOF := False; { keep on in there... }
  end;
  Result := AnsiChar(inBuffer[inBuffPos]);
  InCheckSum:= InChecksum+Ord(Result);
  if EncryptingIn then
    Result := Encrypt(result);
  Inc(inBuffPos);
end;

procedure TCompThreadSafety.PutChar(ch: AnsiChar);
begin
  if MaxNumberOfBytesToWrite <= 0 then { V3.5 }
     raise EUnableToCompress.Create('Could not compress data, will store instead');
  Dec(MaxNumberOfBytesToWrite);

  if outBuffPos >= ChunkSize then { filled buffer }
  begin
    Sleep(0);  { V2.5 replaces Application.ProcessMessages so Compress can be
                  used without Forms dependency }
{$IFDEF CLR}
    dest.WriteBuffer(OutBuffer, ChunkSize);
{$ELSE}
    dest.WriteBuffer(OutBuffer[0], ChunkSize);
{$ENDIF}
    OutBuffPos := 0;
  end;
  if EncryptingOut then
    ch := Encrypt(ch); { really a decrypt }
  OutBuffer[outBuffPos] := Byte(Ch);
  OutCheckSum:= OutChecksum+Ord(ch);
   Inc(OutBuffPos);
   Inc(BytesOut);
end;

{$IFDEF RLE_COMPRESS}
procedure TCompThreadSafety.emit(count: Integer; ch: AnsiChar);
begin
  if (count > 2) or (count=0) then {only emit if worth it }
  begin
    putChar(RLEescapechar);
    putChar(AnsiChar(count));
  end else
  begin
     Dec(count);
     while count > 0 do begin putChar(ch); Dec(count) end;
  end;
end;

procedure TCompThreadSafety.CompressRLE;
var ch: AnsiChar;
begin
 while True do
 begin
   ch:= GetChar;
   if InputEOF then
   begin
     if inRepeat then
        emit(Dupcount,lastch); { flag the repeat }
     break;
   end;
   if inRepeat then
   begin
      if (lastch = ch) and (DupCount<255) then
        Inc(DupCount) { and stay in inRepeat }
      else
      begin
        emit(DupCount,lastch); { however many }
        lastch := ch;
        if ch=RLEescapechar then
        begin
           emit(0,RLEEscapechar); { flag it }
        end else
           Putchar(ch);
        inRepeat := False;
      end;
   end else
   begin
     if (ch=RLEescapechar) then
        emit(0,ch)
     else if (ch=lastch) and not AtStart then
     begin
        DupCount := 2;
        inRepeat := True;
     end else Putchar(ch);
     lastch := ch;
   end;
   AtStart := False;
 end; { While not InputEOF }
end;
{$ENDIF} {RLE_COMPRESS}

{$IFDEF RLE_EXPAND}
procedure TCompThreadSafety.ExpandRLE;
var ch: AnsiChar;
begin
 while True do
 begin
   ch:= GetChar;
   if InputEOF then
     break; { done, at last... }
   if ch <> AnsiChar(RLEescapechar) then
     Putchar(ch)
   else { ok, get a count... MUST be there, really! }
   begin
     DupCount := Ord(GetChar); { 0 if EOF, but not legal, however... }
     if DupCount=0 then Putchar(RLEEscapechar) { special flag }
     else
     begin
        Dec(DupCount);  { because one was already IN the bytestream }
        while Dupcount>0 do begin Putchar(lastch); Dec(DupCount) end;
     end;
   end;
   lastch := ch;
 end; { while }
end;
{$ENDIF} { RLE_EXPAND }

{    Compress/ExpandLZH
     Derived from the Unix C code of the LHarc Encoding/Decoding module
  	LZSS Algorithm			Haruhiko.Okumura
  	Adaptic Huffman Encoding	1989.05.27  Haruyasu.Yoshizaki
     See copyright/distribution permissions at end of the LZH section
     (from the Unix 'man' document accompanying the original file)
}
{	HUFFMAN ENCODING }


{$IFDEF LZH1_COMMON}

{ Encode/Decode table for upper 6 bits position information }
{ for encode }
const p_len: Array [0..63] of AnsiChar =
(	#03, #04, #04, #04, #05, #05, #05, #05,
	#05, #05, #05, #05, #06, #06, #06, #06,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#08, #08, #08, #08, #08, #08, #08, #08,
	#08, #08, #08, #08, #08, #08, #08, #08
);

p_code: Array [0..63] of AnsiChar =
(	AnsiChar($00), AnsiChar($20), AnsiChar($30), AnsiChar($40),
  AnsiChar($50), AnsiChar($58), AnsiChar($60), AnsiChar($68),
	AnsiChar($70), AnsiChar($78), AnsiChar($80), AnsiChar($88),
  AnsiChar($90), AnsiChar($94), AnsiChar($98), AnsiChar($9C),
	AnsiChar($A0), AnsiChar($A4), AnsiChar($A8), AnsiChar($AC),
  AnsiChar($B0), AnsiChar($B4), AnsiChar($B8), AnsiChar($BC),
	AnsiChar($C0), AnsiChar($C2), AnsiChar($C4), AnsiChar($C6),
  AnsiChar($C8), AnsiChar($CA), AnsiChar($CC), AnsiChar($CE),
	AnsiChar($D0), AnsiChar($D2), AnsiChar($D4), AnsiChar($D6),
  AnsiChar($D8), AnsiChar($DA), AnsiChar($DC), AnsiChar($DE),
	AnsiChar($E0), AnsiChar($E2), AnsiChar($E4), AnsiChar($E6),
  AnsiChar($E8), AnsiChar($EA), AnsiChar($EC), AnsiChar($EE),
	AnsiChar($F0), AnsiChar($F1), AnsiChar($F2), AnsiChar($F3),
  AnsiChar($F4), AnsiChar($F5), AnsiChar($F6), AnsiChar($F7),
	AnsiChar($F8), AnsiChar($F9), AnsiChar($FA), AnsiChar($FB),
  AnsiChar($FC), AnsiChar($FD), AnsiChar($FE), AnsiChar($FF)
);

{ for decode }
d_code: Array [0..255] of AnsiChar =
(	#00, #00, #00, #00, #00, #00, #00, #00,
	#00, #00, #00, #00, #00, #00, #00, #00,
	#00, #00, #00, #00, #00, #00, #00, #00,
	#00, #00, #00, #00, #00, #00, #00, #00,
	#01, #01, #01, #01, #01, #01, #01, #01,
	#01, #01, #01, #01, #01, #01, #01, #01,
	#02, #02, #02, #02, #02, #02, #02, #02,
	#02, #02, #02, #02, #02, #02, #02, #02,
	#03, #03, #03, #03, #03, #03, #03, #03,
	#03, #03, #03, #03, #03, #03, #03, #03,
	#04, #04, #04, #04, #04, #04, #04, #04,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#08, #08, #08, #08, #08, #08, #08, #08,
	#09, #09, #09, #09, #09, #09, #09, #09,
	AnsiChar($0A), AnsiChar($0A), AnsiChar($0A), AnsiChar($0A),
  AnsiChar($0A), AnsiChar($0A), AnsiChar($0A), AnsiChar($0A),
	AnsiChar($0B), AnsiChar($0B), AnsiChar($0B), AnsiChar($0B),
  AnsiChar($0B), AnsiChar($0B), AnsiChar($0B), AnsiChar($0B),
	AnsiChar($0C), AnsiChar($0C), AnsiChar($0C), AnsiChar($0C),
  AnsiChar($0D), AnsiChar($0D), AnsiChar($0D), AnsiChar($0D),
	AnsiChar($0E), AnsiChar($0E), AnsiChar($0E), AnsiChar($0E),
  AnsiChar($0F), AnsiChar($0F), AnsiChar($0F), AnsiChar($0F),
	AnsiChar($10), AnsiChar($10), AnsiChar($10), AnsiChar($10),
  AnsiChar($11), AnsiChar($11), AnsiChar($11), AnsiChar($11),
	AnsiChar($12), AnsiChar($12), AnsiChar($12), AnsiChar($12),
  AnsiChar($13), AnsiChar($13), AnsiChar($13), AnsiChar($13),
	AnsiChar($14), AnsiChar($14), AnsiChar($14), AnsiChar($14),
  AnsiChar($15), AnsiChar($15), AnsiChar($15), AnsiChar($15),
	AnsiChar($16), AnsiChar($16), AnsiChar($16), AnsiChar($16),
  AnsiChar($17), AnsiChar($17), AnsiChar($17), AnsiChar($17),
	AnsiChar($18), AnsiChar($18), AnsiChar($19), AnsiChar($19),
  AnsiChar($1A), AnsiChar($1A), AnsiChar($1B), AnsiChar($1B),
	AnsiChar($1C), AnsiChar($1C), AnsiChar($1D), AnsiChar($1D),
  AnsiChar($1E), AnsiChar($1E), AnsiChar($1F), AnsiChar($1F),
	AnsiChar($20), AnsiChar($20), AnsiChar($21), AnsiChar($21),
  AnsiChar($22), AnsiChar($22), AnsiChar($23), AnsiChar($23),
	AnsiChar($24), AnsiChar($24), AnsiChar($25), AnsiChar($25),
  AnsiChar($26), AnsiChar($26), AnsiChar($27), AnsiChar($27),
	AnsiChar($28), AnsiChar($28), AnsiChar($29), AnsiChar($29),
  AnsiChar($2A), AnsiChar($2A), AnsiChar($2B), AnsiChar($2B),
	AnsiChar($2C), AnsiChar($2C), AnsiChar($2D), AnsiChar($2D),
  AnsiChar($2E), AnsiChar($2E), AnsiChar($2F), AnsiChar($2F),
	AnsiChar($30), AnsiChar($31), AnsiChar($32), AnsiChar($33),
  AnsiChar($34), AnsiChar($35), AnsiChar($36), AnsiChar($37),
	AnsiChar($38), AnsiChar($39), AnsiChar($3A), AnsiChar($3B),
  AnsiChar($3C), AnsiChar($3D), AnsiChar($3E), AnsiChar($3F)
);

d_len: Array [0..255] of AnsiChar =
(	#03, #03, #03, #03, #03, #03, #03, #03,
	#03, #03, #03, #03, #03, #03, #03, #03,
	#03, #03, #03, #03, #03, #03, #03, #03,
	#03, #03, #03, #03, #03, #03, #03, #03,
	#04, #04, #04, #04, #04, #04, #04, #04,
	#04, #04, #04, #04, #04, #04, #04, #04,
	#04, #04, #04, #04, #04, #04, #04, #04,
	#04, #04, #04, #04, #04, #04, #04, #04,
	#04, #04, #04, #04, #04, #04, #04, #04,
	#04, #04, #04, #04, #04, #04, #04, #04,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#05, #05, #05, #05, #05, #05, #05, #05,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#06, #06, #06, #06, #06, #06, #06, #06,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#07, #07, #07, #07, #07, #07, #07, #07,
	#08, #08, #08, #08, #08, #08, #08, #08,
	#08, #08, #08, #08, #08, #08, #08, #08
);

{ Initialize tree }

procedure TCompThreadSafety.StartHuff;
var i,j: Smallint;
begin
  for i := 0 to N_CHAR-1 do
  begin
    freq[i] := 1;
    son[i] := i + constT;
    prnt[i + constT] := i;
  end;
  i := 0;
  j := N_CHAR;
  while (j <= constR) do
  begin
    freq[j] := freq[i]+freq[i + 1];
    son[j] := i;
    prnt[i] := j;
    prnt[i + 1] := j;
    Inc(i,2);
    Inc(j);
  end;
  freq[constT] := $ffff;
  prnt[constR] := 0;
  putlen := #0;
  getlen := #0; { good to see... }
  putbuf := 0;
  getbuf := 0;
end;

procedure TCompThreadSafety.EndHuff;
begin
  // V8.0 - Used to be FreeMem for buffers - now obsolete
end;


{ reconstruct tree }
procedure TCompThreadSafety.reconst;
var i,j,k: Smallint;
    f: Word;
    pIdx, eIdx: Integer;
    piIdx, eiIdx: Integer;
begin
	{ correct leaf node into of first half,
	   and set these freqency to (freq+1)/2  }
  j := 0;
  for i := 0 to constT-1 do
  begin
    if (son[i] >= constT) then
    begin
      freq[j] := (freq[i] + 1) div 2;
      son[j] := son[i];
      Inc(j);
    end;
  end;
	{ build tree.  Link sons first }

  i:=0;
  for J := N_CHAR to constT-1 do
  begin
    k := i + 1;
    f := freq[i] + freq[k];
    freq[j] := f;
    k := j - 1;
    while f < freq[k] do
      Dec(k);
    inc(k);
    pIdx := j;
    eIdx := k;
    while pIdx > eIdx do
    begin
      freq[pIdx] := freq[pIdx-1];
      Dec(pIdx);
    end;
    freq[k] := f;
    piIdx := j;
    eiIdx := k;
    while piIdx > eiIdx do
    begin
      son[piIdx] := son[piIdx-1];
      Dec(piIdx);
    end;
    son[k] := i;
    Inc(i,2);
  end;
	{ link parents }
  for i:= 0 to constT-1 do
  begin
    k:= son[i];
    if (k >= constT) then
      prnt[k] := i
    else
    begin
      prnt[k] := i;
      prnt[k + 1] := i;
    end;
  end;
end;

{ update given code's frequency, and update tree }

procedure TCompThreadSafety.update (c: Word);
var i,j,l: Smallint;
    k: Word; { was Smallint, caused overflow }
begin
  if (freq[constR] = MAX_FREQ) then
    reconst;
  c := prnt[c + constT];

  while true do
  begin
     Inc(freq[c]);
     k := freq[c];

	{ swap nodes when the frequency order is no longer correct }
     l := c+1;
     if (k > freq[l]) then
     begin
      Inc(l);
      while k > freq[l] do
        Inc(l);
      dec(l);
      freq[c] := freq[l];
      freq[l] := k;
      i := son[c];
      prnt[i] := l;
      if (i < constT) then
        prnt[Succ(i)] := l;

      j := son[l];
      son[l] := i;

      prnt[j] := c;
      if (j < constT) then
        prnt[Succ(j)] := c;
      son[c] := j;

      c := l;
    end;
    c:= prnt[c];
    if c=0 then
      break; { all done  -- at root }
  end;
end;

{$ENDIF} {LZH1_COMMON}

{ Compression-only routines }

{$IFDEF LZH1_COMPRESS}

procedure TCompThreadSafety.InitTree;
var i:Smallint;
begin
  for i := constN+1 to constN+constN do rson[i] := TNIL;
  for i := 0 to constN-1 do dad[i] := TNIL;
end;

{ Insert to node }
procedure TCompThreadSafety.InsertNode(r: Smallint);
var cmp, p: Smallint;
     c,i,j: Word;
     keyIdx: Integer;
begin
  rson[r] := TNIL;
  lson[r] := TNIL;
  keyIdx := r;
  i := Smallint(text_buf[keyIdx+1]) xor Smallint(text_buf[keyIdx+2]);
  i := i xor (i shr 4);
  p := constN + 1 + Smallint(text_buf[keyIdx]) + ((i and $0f) shl 8);
  cmp := 1;
  i := 1;
  j := 1;
  match_length := 0;

  while True do
  begin
    if (cmp >= 0) then
    begin
    	if (rson[p] <> TNIL) then
        begin
          p := rson[p];
	  j := Ord(same[p]);
	end else begin
	  rson[p] := r;
	  dad[r] := p;
	  same[r] := AnsiChar(i);
	  exit;
        end;
    end else begin
    	if (lson[p] <> TNIL) then
        begin
	  p := lson[p];
	  j := Ord(same[p]);
	end else begin
	  lson[p] := r;
	  dad[r] := p;
	  same[r] := AnsiChar(i);
	  exit;
        end;
     end;

     if (i > j) then
     begin
       i := j;
       cmp := Ord(text_buf[keyIdx+i]) - Ord(text_buf[p + i]);
     end else
       if (i = j) then
       begin
         while i < constF do
         begin
     	   cmp := Ord(text_buf[keyIdx+i]) - Ord(text_buf[p + i]);
  	   if (cmp <> 0) then
              break;
           Inc(i);
         end; { while }
       end;

       if (i > THRESHOLD) then
       begin
       	 if (i > match_length) then
         begin
	    match_position := ((r - p) and (constN - 1)) - 1;
            match_length := i;
	    if (match_length  >= constF) then
               break
	 end else if (i = match_length) then
         begin
            c:= ((r - p) and (constN - 1)) - 1;
  	    if (c < match_position) then
  	    	match_position := c;
  	 end
      end { if i > THRESHOLD }
  end; { while true }

  same[r] := same[p];
  j := dad[p];
  dad[r] := j;
  lson[r] := lson[p];
  rson[r] := rson[p];
  dad[lson[p]] := r;
  dad[rson[p]] := r;
  if (rson[j] = p) then
    rson[j] := r
  else
    lson[j] := r;
  dad[p]:= TNIL;
end;

procedure TCompThreadSafety.link (n, p, q: Smallint);
var s1Idx, s2Idx, s3Idx: Integer;
begin
  if (p >= TNIL) then begin
    same[q] := #1;
    exit;
  end;

  s1Idx := p + n;
  s2Idx := q + n;
  s3Idx := p + constF;
  while (s1Idx < s3Idx) do
  begin
    if text_buf[s1Idx] <> text_buf[s2Idx] then
    begin
      same[q] := AnsiChar(s1Idx-p);
      exit;
    end;
    inc(s1Idx);
    inc(s2Idx);
  end;
  same[q] := AnsiChar(constF);
end;

procedure TCompThreadSafety.linknode (p, q, r: smallint);
var cmp: Smallint;
begin
  cmp := Ord(same[q]) - Ord(same[r]);
  if (cmp = 0) then
    link(Ord(same[q]), p, r)
  else if (cmp < 0) then
    same[r] := same[q];
end;

procedure TCompThreadSafety.DeleteNode(p: Smallint);
var q: Smallint;
begin
  if (dad[p] = TNIL) then
    exit;{ nothing linked }

  if (rson[p] = TNIL) then
  begin
     q:=lson[p];
     if (q <> TNIL) then
  	linknode(dad[p], p, q);
  end else if (lson[p] = TNIL) then
  begin
     q := rson[p];
     linknode(dad[p], p, q);
  end else
  begin
     q := lson[p];
     if (rson[q] <> TNIL) then
     begin
        q:=rson[q];
        while rson[q]<>TNIL do
      	  q := rson[q];
  	if (lson[q] <> TNIL) then
  	   linknode(dad[q], q, lson[q]);
	link(1, q, lson[p]);
	rson[dad[q]] := lson[q];
	dad[lson[q]] := dad[q];
	lson[q] := lson[p];
  	dad[lson[p]] := q;
     end;
     link(1, dad[p], q);
     link(1, q, rson[p]);
     rson[q] := rson[p];
     dad[rson[p]] := q;
   end;
   dad[q] := dad[p];
   if (rson[dad[p]] = p) then
     rson[dad[p]] := q
   else
     lson[dad[p]] := q;
   dad[p] := TNIL;
end;

{ output C bits }
procedure TCompThreadSafety.Putcode(L: Smallint; c: Word);
var len: smallint;
    b: Word;
begin

  len := Ord(putlen);
  b := putbuf;
  b := b or (c shr len);
  Inc(len,L);
  if (len>= 8) then
  begin
    putchar(AnsiChar(b shr 8));
    Dec(len,8);
    if (len >= 8) then
    begin
      putchar(AnsiChar(b));
      Dec(len,8);
{$R-}  { Always }
      b := c shl (L-len);
{$R-}  { R+ if doing range check during debug }
    end else
      b:= b shl 8;
  end;
  putbuf := b;
  putlen := AnsiChar(len);
end;

procedure TCompThreadSafety.EncodeChar (c:word);
var	i: Longint;  { careful -- should be UNSIGNED long! }
  	k,j : Smallint;
begin
  i := 0;
  j := 0;
  k := prnt[c+constT];

	{ trace links from leaf node to root }
  while true do
  begin
    i := i shr 1;
	{ if node index is odd, trace larger of sons }
    if (k and 1) >0 then
        Inc(i,$80000000);
    Inc(j);
    k:=prnt[k];
    if k=constR then break;
  end;

  if (j > 16) then
  begin
    Putcode(16, Word(i shr 16));
    Putcode(j - 16, word(i));
  end else
    Putcode(j, Word(i shr 16));

  update(c);
end;

procedure TCompThreadSafety.EncodePosition(c:Word);
var i: Word;
begin
  { output upper 6 bits from table }
   i := c shr 6;
   Putcode(Smallint(p_len[i]), Word(p_code[i]) shl 8);
  { output lower 6 bits }
   Putcode(6, Word(c and $3f) shl 10);
end;

procedure TCompThreadSafety.EncodeEnd;
begin
  if (putlen>#0) then
     putchar(AnsiChar(putbuf shr 8));
end;

procedure TCompThreadSafety.CompressLZH;
var i, r, s, last_match_length: smallint;
    len: word; { was smallint -- this silences D4 warning about signed/unsigned compare }
    c: AnsiChar;
begin
  try
    StartHuff;
    InitTree;
    s := 0;
    r := constN - constF;

    for i:= s to r-1 do text_buf[i] := ' ';

    for len:=0 to constF-1 do
    begin
      c:=getchar;
      if InputEOF then
        break;
      text_buf[r + len] := c;
    end;
    if len=0 then
      exit; { added PJH }
    if not InputEOF then
      len:=constF; { so we have correct value for finishing up... }

    for i := 1 to constF do InsertNode(r - i);
       InsertNode(r);

    while true do
    begin
      if (match_length > len) then
        match_length := len;
      if (match_length <= THRESHOLD) then
      begin
        match_length := 1;
        EncodeChar(Word(text_buf[r]));
      end else
      begin
        EncodeChar(255 - THRESHOLD + match_length);
        EncodePosition(match_position);
      end;
      last_match_length := match_length;
      i:=0;
      while i < last_match_length do
      begin
        c:= getChar;
        if InputEof then
          break;
        DeleteNode(s);
        text_buf[s] := c;
        if (s < constF - 1) then
          text_buf[s + constN] := c;
        s := (s + 1) and (constN - 1);
        r := (r + 1) and (constN - 1);
        InsertNode(r);
        Inc(i);
      end;

      while i < last_match_length do
      begin
        Inc(i);
        DeleteNode(s);
        s := (s + 1) and (constN - 1);
        r := (r + 1) and (constN - 1);
        Dec(len);
        if len > 0 then
        begin
          InsertNode(r);
        end;
      end; { don't need the extra Inc(i) as not used }
      if len=0 then break;
    end;
    EncodeEnd;
  finally
    EndHuff;
  end;
end;
{$ENDIF} { LZH1_COMPRESS }


{ Routines used only for LZH1 expansion }

{$IFDEF LZH1_EXPAND}
{ get one bit
 returning in Bit 0 }
function TCompThreadSafety.GetBit: Smallint;
var dx, c: Word;
begin
{$R-} { Rangechecks off in case on...}
  dx := getbuf;

  if (getlen <= #8) then
  begin
    c := Word(getchar);
    if InputEOF then c := 0;
    dx := dx or (c shl (8-Ord(getlen)));
    Inc(getlen,8);
  end;
  getbuf := dx shl 1;
  Dec(getlen);
  if (dx and $8000) <>0 then
     result:=1
  else
     result := 0;
end;

{get one byte
 returning in Bits 7..0 }
function TCompThreadSafety.GetByte: Word;
var dx, c: Word;
begin
  dx := getbuf;

  if (getlen <= #8) then
  begin
    c := Word(getchar);
    if InputEOF then c := 0;
    dx := dx or (c shl (8-Ord(getlen)));
    Inc(getlen,8);
  end;
  getbuf := dx shl 8;
  Dec(getlen,8);
  result := (dx shr 8) and $ff;
end;


{ get N bits
{ returning in Bit(n-1)...Bit 0 }
const mask: Array[0..16] of Word =
   ( $0000,
     $0001, $0003, $0007, $000f,
     $001f, $003f, $007f, $00ff,
     $01ff, $03ff, $07ff, $0fff,
     $1fff, $3fff, $0fff, $ffff );

     shift: Array[0..16] of Smallint =
     	( 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0 );

function TCompThreadSafety.GetNBits(n: Word): Smallint;
var dx, c: Word;
begin
  dx := getbuf;
  if (getlen <= #8) then
  begin
    c := Word(getchar);
    if InputEOF then c := 0;
    dx := dx or (c shl (8-Ord(getlen)));
    Inc(getlen,8);
  end;
  getbuf := dx shl n;
  Dec(getlen,n);
  result := (dx shr shift[n]) and mask[n];
end;

function TCompThreadSafety.DecodeChar: Smallint;
begin
  result := son[constR];
 { trace from root to leaf
   got bit is 0 to small(son[]), 1 to large (son[]+1) son node }
  while (result < constT) do
  begin
    if Getbit > 0 then Inc(result);
    result := son[result];
  end;
  Dec(result,constT);
  update(result);
end;

function TCompThreadSafety.DecodePosition: Smallint;
var i,j,c: Word;
begin
  { decode upper 6 bits from table }
  i := GetByte;
  c := Word(d_code[i]) shl 6;
  j := Word(d_len[i]);

  { get lower 6 bits }
  Dec(j,2);
  result := c or  ((( i shl j) or GetNBits(j)) and $3f);
end;

procedure TCompThreadSafety.ExpandLZH(textsize: Int64);  { V8.0 upgraded from Longint }
var	i, j, k, r, c: Smallint;
	count: Int64;  { V8.0 upgraded from Longint }
begin
  if (textsize = 0) then
    exit;

  try
    StartHuff;
    r := constN - constF;
    for i:= 0 to r-1 do
      text_buf[i] := ' ';
    count:=0;
    while count < textsize do
    begin
      c := DecodeChar;
      if (c < 256) then
      begin
        Putchar(AnsiChar(c));
        text_buf[r] := AnsiChar(c);
        Inc(r);
        r := r and (constN - 1);
        Inc(count);
      end else
      begin
        i := (r - DecodePosition - 1) and (constN - 1);
        j := c - 255 + THRESHOLD;
        for k := 0 to j-1 do
        begin
          c := Ord(text_buf[(i + k) and (constN - 1)]);
          Putchar(AnsiChar(c));
          text_buf[r] := AnsiChar(c);
          inc(r);
          r := r and (constN - 1);
          Inc(count);
        end; {for }
      end;   {else }
    end;      {while }
  finally
    EndHuff;
  end;
end;
{$ENDIF} { LZH1_EXPAND }

{
October 6, 1990   INTEGRATED SOLUTIONS 4.3 BSD                  4

LHARC(L)            UNIX Programmer's Manual             LHARC(L)

DISTRIBUTION
     This software is Copyrighted 09/19/1989 by Yooichi Tagawa,
     Tokyo, Japan.  The author allows free distribution under the
     following conditions:

     1.      This copyright notice is maintained.

     2.      As for contents of distribution,

             a.      All files in the standard distribution (e.g.
                     source code, documents, guide to program-
                     mers) must exist in any re-distribution.  If
                     they were modified, modifications must be
                     clearly indicated.

             b.      If any kind of valuable capabilities, pro-
                     grams, or information were added within a
                     re-distribution, the distributor doing the
                     modifications must make every effort to
                     include them in the next distribution.  In
                     such case, decent documents describing the
                     new programs or additional capabilities must
                     be prepared.

             c.      Do not distribute LHarc in binary only.
                     (This also pertains to added programs or
                     information).

     3.      You must make an effort to distribute the newest
             version (but it's not your mandatory duty).

     4.      This program comes with no warranty, expressed or
             implied.  Use of this program is at the entire risk
             of the user.

     5.      The current author has no duty to correct mistakes,
             if any, in this program.

     6.      Parts of this program, or the whole program may be
             freely incorporated into other programs.  In such
             case, the resulting program is not allowed to be
             called LHarc.

     7.      For use of this program for commercial purposes, in
             addition to the above conditions, the following con-
             ditions also apply.

             a.      Commercial use which consists solely or
                     principally of this program as a product for
                     sale, is prohibited.

             b.      If the purchaser or licensee of a commercial
                     product including LHarc in its distribution
                     is judged likely to make improper use of
                     this program, or refuses to comply with
                     these restrictions, distribution of LHarc
                     with the commercial product to that user is
                     forbidden.

             c.      If LHarc is used as a means of installing
                     commercial software, the vendor of the com-
                     mercial software is not allowed to require
                     the user to do that installation.  In this
                     case, the commercial vendor must do the
                     installation.  All damages caused by use of
                     LHarc in such an installation are the
                     responsibility of the commercial vendor.

             d.      If LHarc is included as part of a commercial
                     product, the vendor of that product must
                     support this program.

ACKNOWLEDGEMENTS
     Thanks to Haruhiko Okumura for his development of the LZARI
     method, on which LZHUF was based.  Thanks to Haruhiko Miki
     for LArc, and thanks to Yasuaki Yoshizaki for his derivative
     work from LArc, thus making up the LZHUF method and MS-DOS
     LHarc. Thanks to the many people who assisted the develop-
     ment of LHarc, Kazuaki Ishibashi for forwarding mail
     (between commercial BBS's), all beta testers who reported
     various results, and the person who forwarded this informa-
     tion, Koichiro Mori.

     This manual was translated by Youki Kadobayashi
     <youki@mercury.ics.osaka-u.ac.jp>.  Although I made every
     possible effort I could, there may be mistakes or queer
     expressions.  Feel free to correct my mistakes, I will
     appreciate your feedback.
}

const CHAR_BIT = 8; { special for LZH5 in TCompress }

{ Now for the LZH5 handling routines V3.5 }

{$IFDEF LZH5_COMMON}

{$IFDEF COMPARE_BYTES}
procedure TCompThreadSafety.chkchar(c: smallint);
var inch: AnsiChar;
begin
  if compareFile.read(inch,1)<>1 then
   raise Exception.Create('EOF on compare file!');
  if inch<> AnsiChar(c) then
   raise Exception.Create('bad compare byte: '+InttoHex(ord(inch),2)+' vs. '+InttoHex(c,2)+' at '+IntToStr(BytesOut));
end;
{$ENDIF}

procedure TCompThreadSafety.StartHuff5(Compress: Boolean); { init bit get/put vars }
begin { also allocate and init internal structures }
  if Compress then
  begin
    Bitcount := AnsiChar(Char_Bit);
  end else
  begin
    Bitcount := #0;
  end;

  depth := #0;
  subBitbuf := #0;
  bitbuf := 0;
  blocksize := 0;
end;

procedure TCompThreadSafety.EndHuff5;
begin
  // V8.0 - Used to be FreeMem for buffers - now obsolete
end;
{$ENDIF} { LZH5_COMMON}

{$IFDEF LZH5_EXPAND}
procedure TCompThreadSafety.buildtable(nchar: Smallint; bitlen: Array of AnsiChar;
                            tablebits: Smallint; var table: Array of Word);
type
   TPtrType = (ptLeft, ptRight, ptTable);

var count: array[0..16] of Word; { rightsized }
    weight: array[0..16] of Word;  { 0 unused... }
    start: array[0..16] of Word;
    total, i: Word;
    j, k, l, m, n, avail: Longint;
    pType: TPtrType;
    pIndex, pValue: Word;
begin
   { initialise }
   avail := nchar;
   for i := 1 to 16 do
   begin
     count[i] := 0;
     weight[i] := 1 shl (16-i);
   end;

   for i := 0 to pred(nchar) do
      inc(count[ord(bitlen[i])]); { count[bitlen[i]] := couint[bitlen[i]]+1; }

   total := 0;
   for i := 1 to 16 do
   begin
     start[i] := total;
     inc(total,weight[i]*count[i]);
   end;
   if (total and $ffff) <> 0 then
     raise ELZHBadDecodeTable.Create('Bad decode table in archive!');

   m := 16 - tablebits;
   for i := 1 to tablebits do
   begin
     start[i] := start[i] shr m;
     weight[i] := weight[i] shr m;
   end;

   j := start[tablebits+1] shr m;
   k := 1 shl tablebits;
   if j <>0 then
     for i := j to pred(k) do
       table[i] := 0;

  for j := 0 to pred(nchar) do
  begin
    k := Longint(bitlen[j]);
    if k=0 then continue;
    l := start[k] + weight[k];

    if k <= tablebits then
    begin
      for i := start[k] to pred(l) do
        table[i] := j;
    end else begin
      i := start[k];
      ptype := pttable;
      pindex := i shr m;
      pvalue := table[pindex];
      i := (i shl tablebits) and $FFFF;   { and $FFFF - from David}

      for n := 0 to pred(k - tablebits) do
      begin
        if pValue = 0 then
        begin
          right[avail] := 0;
          left[avail] := 0;
          case pType of
            ptLeft:  left[pindex]  := avail;
            ptRight: right[pindex] := avail;
            ptTable: table[pindex] := avail;
          end;
          pvalue := avail;
          inc(avail);
        end; { if }
        if (i and $8000) <> 0 then { V4.0 change >0 to <> 0 }
        begin
          pIndex := pValue;
          pType := ptRight;
          pValue := right[pindex];
        end else begin
          pIndex := pValue;
          pType := ptLeft;
          pValue := left[pindex];
        end;
        i := (i shl 1) and $FFFF;  { and $FFFF - from David}
      end; { while }
      case pType of
        ptLeft:  left[pindex]  := j;
        ptRight: right[pindex] := j;
        ptTable: table[pindex] := j;
      end;
    end; { else }
    start[k] := l and $FFFF;   { and $FFFF - from David}
  end; { for }
end;

procedure TCompThreadSafety.fillbuf(n: AnsiChar);
begin
  while (n > bitcount) do
  begin
    Dec(n,ord(bitcount));
    bitbuf := (bitbuf shl ord(bitcount)) + (ord(subbitbuf) shr (Char_Bit - ord(bitcount)));
    if InputEOF then
       subBitBuf := #0
    else
       subBitBuf := GetChar;
    bitcount := AnsiChar(Char_bit);
  end;
  Dec(bitcount,ord(n));
  bitbuf := (bitbuf shl ord(n)) +(ord(subBitBuf) shr (Char_bit - ord(n)));
  subbitbuf :=   AnsiChar(ord(subbitbuf) shl ord(n));
end;

function TCompThreadSafety.getbits(n: AnsiChar): Word;
begin
   result := bitbuf shr (2*Char_Bit - ord(n));
   fillbuf(n)
end;

procedure TCompThreadSafety.read_position_len(nn, nbit, i_special: Smallint);
var i,c,n: Smallint;
    mask: Word;
begin
   n := getbits(AnsiChar(nbit));
   if n=0 then
   begin
     c := getbits(AnsiChar(nbit));
     for i := 0 to pred(nn) do
      position_len[i] := #0;
     for i := 0 to pred(256) do
      position_table[i] := c;
   end else
   begin
     i := 0;
     while i < n do
     begin
       c := bitbuf shr (16-3);
       if c=7 then
       begin
         mask := 1 shl (16-4);
         while (mask and bitbuf) > 0 do
         begin
           mask := mask shr 1;
           inc(c);
         end;
       end; {if }
       if c < 7 then
         fillbuf(#3)
       else
         fillbuf(AnsiChar(c-3));
       position_len[i] := AnsiChar(c); inc(i);
       if i = i_special then
       begin
         for c := pred(getbits(#2)) downto 0 do
         begin
           position_len[i] := #0; inc(i);
         end;

       end; { if }
     end; { while }
     while (i < nn) do
     begin
       position_len[i] := #0; inc(i);
     end;
     buildtable(nn,position_len,8,position_table);
   end; { else }
end;

procedure TCompThreadSafety.read_charLen;
var i,c,n: Smallint;
    mask: Word;
begin
   n := getbits(AnsiChar(CBIT));
   if n=0  then
   begin
     c := getbits(AnsiChar(CBIT));
     for i:= 0 to pred(constNC) do
       charLen[i] :=#0;
     for i := 0 to pred(4096) do
       char_table[i] := c;
   end else begin
     i := 0;
     while i < n do
     begin
       c := position_table[bitbuf shr (16-8)];
       if c >= constNT then
       begin
          mask := 1 shl (16-9);
          repeat
            if (bitbuf and mask) > 0 then
               c := right[c]
            else
               c := left[c];
            mask := mask shr 1;
          until c < constNT;
       end;
       fillbuf(position_len[c]);
       if c <= 2 then
       begin
         if c=0 then
            c := 1
         else if c = 1 then
            c := getbits(#4)+3
         else
            c := getbits(AnsiChar(CBIT))+20;
         for c := 0 to pred(c) do
         begin
            charLen[i] := #0;
            inc(i);
         end;
       end else
       begin
          charLen[i] := AnsiChar(c-2);
          inc(i);
       end; { else }
     end; { while }

     while (i < constNC) do
     begin
       charLen[i] := #0;
       inc(i);
     end;
     buildtable(constNC,charLen,12,char_table);
   end; { else }
end;


function TCompThreadSafety.DecodeChar5: Word;
var mask: Word;
begin
   if blocksize =0 then
   begin
     blocksize := getbits(#16);
     read_position_len(constNT,TBIT,3);
     read_charLen;
     read_position_len(constNP,PBIT,-1);
   end;
   dec(blocksize);
   result := char_table[bitbuf shr 4];
   if result < constNC then
      fillbuf(charLen[result])
   else begin
      fillbuf(#12);
      mask := 1 shl (16-1);
      repeat
        if (bitbuf and mask) > 0 then
           result := right[result]
        else
           result := left[result];
        mask := mask shr 1;
      until result < constNC;
      fillbuf(AnsiChar(ord(charLen[result]) - 12));
   end; { else }
end;

function TCompThreadSafety.DecodePosition5: Word;
var mask: Word;
begin
  result := position_table[bitbuf shr (16-8)];
  if result < constNP then
    fillbuf(position_len[result])
  else begin
    fillbuf(#8);
    mask := 1 shl (16-1);
    repeat
      if (bitbuf and mask) > 0 then
        result := right[result]
      else
        result := left[result];
      mask := mask shr 1;
    until result < constNP;
    fillbuf(AnsiChar(ord(position_len[result])-8));
  end; { else }
  if result <> 0 then
    result := (1 shl pred(result)) + getbits(AnsiChar(pred(result)));
end;


procedure TCompThreadSafety.ExpandLZH5(destsize: Int64);  { V8.0 upgraded from Longint }
var i, j, k, c, offset: Smallint;
    count: Int64;  { V8.0 upgraded from Longint }
    loc: Word;
begin
if destsize=0 then exit; { quietly ignore }

try
{$IFDEF COMPARE_ON_EXPAND}
     CompareFile := TFileStream.create('c:\x',fmOpenRead);
{$ENDIF}


 StartHuff5(False); { initialise important things }
 fillbuf(AnsiChar(2*CHAR_BIT)); { prime input buffer }

 for i:= 0 to pred(DictionarySize) do
  text[i] := ' ';

 offset := 256-3;
 count := 0;
 loc := 0;
 while count < DestSize do
 begin
   c := DecodeChar5;
   if (c <= CharMax) then
   begin
     text[loc] := AnsiChar(c); inc(loc);
     PutChar(AnsiChar(c));
{$IFDEF COMPARE_ON_EXPAND}
     chkchar(c);
{$ENDIF}
     if (loc=DictionarySize) then
       loc := 0;
     inc(count);
   end else
   begin
      j := c - offset;
      i := (loc - DecodePosition5 -1 ) and pred(DictionarySize);
      inc(count,j);
      for k:=0 to pred(j) do
      begin
        text[loc] := text[(i+k) and pred(DictionarySize)];
        PutChar(text[loc]);
{$IFDEF COMPARE_ON_EXPAND}
     chkchar(ord(text[loc]));
{$ENDIF}
        inc(loc);

        if (loc=DictionarySize) then
          loc := 0;
      end { for }
   end; { else }
 end;

finally
{$IFDEF COMPARE_ON_EXPAND}
  CompareFile.free;
{$ENDIF}

  EndHuff5; { free up stuff }
end; { try }
end; { expandLZH5 }

{$ENDIF} { LZH5_EXPAND}



{ Compress routines follow }
{$IFDEF LZH5_COMPRESS}

function TCompThreadSafety.buildtree(nparm: Smallint; var freqparm: Array of Word; var lenparm: Array of AnsiChar;
     var codeparm: Array of Word): Smallint;
var n, heapsize: smallint;
    heap: Array[0..constNC] of Smallint;
    sortIdx: Integer;
    len_cnt: Array[0..16] of Smallint;

    i,j,avail: Smallint; { only used within make-tree }

  procedure buildcode(n: Smallint; len: Array of AnsiChar; var code: Array of Word);
  var weight: Array[0..16] of Word;
      start: Array[0..16] of Word;
      i, j, k: Word;
  begin
     j:=0;
     k:= 1 shl (16-1);
     for i:= 1 to 16 do
     begin
          start[i] := j;
          weight[i] := k;
          inc(j,k* len_cnt[i]);
          k := k shr 1;
     end;
     for i := 0 to pred(n) do
     begin
          j := ord(len[i]);
          code[i] := start[j];
          inc(start[j],weight[j]);
     end;
  end; { buildcode }

  procedure count_len(i: Smallint); { i being the root }
  begin
    if i < n then
    begin
       if depth < #16 then
          inc(len_cnt[ord(depth)])
       else
          inc(len_cnt[16]);
    end else
    begin
       inc(depth);
       count_len(left[i]);
       count_len(right[i]);
       dec(depth);
    end;
  end; { count_len }

  procedure buildlen(root: Smallint);
  var i, k: Smallint;
      cum: Word;
  begin

    for i:=0 to 16 do
      len_cnt[i] := 0;
    count_len(root);
    cum :=0;
    for i:=0 to 16 do
      inc(cum,len_cnt[i] shl (16-i));

    if cum > 0 then { adjust so WILL be zero... }
    begin
      Dec(len_cnt[16],cum);
      repeat
        for i := 15 downto 1 do
          if len_cnt[i] > 0 then
          begin
            dec(len_cnt[i]);
            inc(len_cnt[succ(i)],2);
            break;
          end;
        dec(cum);
      until cum = 0;
    end;

    for i := 16 downto 1 do
    begin
      for k := len_cnt[i] downto 1 do
      begin
        lenparm[codeparm[sortIdx]] := AnsiChar(i); 
        inc(sortIdx);
      end;
    end;

  end; { buildlen }

  procedure downheap(i: Smallint);
  var J,k: Smallint;
  begin
    k := heap[i];
    while True do
    begin
      j := 2*i;
      if j > heapsize then
         break;
      if (j < heapsize) and (freqparm[heap[j]] > freqparm[heap[succ(j)]]) then
         inc(j);
      if freqparm[k] <= freqparm[heap[j]] then
         break;
      heap[i] := heap[j];
      i := j;
    end;
    heap[i] := k;
  end; { downheap }

begin { buildtree }
  n := nparm;
  avail := n;
  heapsize := 0;
  heap[1] := 0;
  for i := 0 to pred(n) do
  begin
    lenparm[i] := #0;
    if freqparm[i] <> 0 then
    begin
      inc(heapsize);
      heap[heapsize] := i;
    end;
  end;

  if heapsize < 2 then
  begin
    codeparm[heap[1]] := 0;
    result := heap[1];
    exit;
  end;

  for i := heapsize div 2 downto 1 do
    downheap(i);

  sortIdx := 0;
  repeat
    i := heap[1];
    if i < n then
    begin
      codeparm[sortIdx] := i;
      inc(sortIdx);
    end;
    heap[1] := heap[heapsize]; dec(heapsize);
    downheap(1);
    j := heap[1];
    if j < n then
    begin
      codeparm[sortIdx] := j;
      inc(sortIdx);
    end;
    result := avail;
    inc(avail);
    freqparm[result] := freqparm[i]+ freqparm[j];
    heap[1] := result;
    downheap(1);
    left[result] := i;
    right[result] := j;
  until heapsize = 1;
  sortIdx := 0;
  buildlen(result);
  buildcode(nparm,lenparm,codeparm);
end; { buildtree }

procedure TCompThreadSafety.count_t_freq;
var i, k, n, count: Smallint;
begin
  for i := 0 to pred(constNT) do
      t_freq[i] :=0;
  n := constNC;
  while (n > 0) and (charLen[pred(n)] = #0) do
     dec(n);
  i := 0;
  while i < n do
  begin
    k := Ord(charLen[i]); inc(i);
    if k = 0 then
    begin
      count := 1;
      while (i < n) and (charLen[i] = #0) do
      begin
        inc(i);
        inc(count);
      end;
      if count <=2 then
         inc(t_freq[0],count)
      else if count <= 18 then
         inc(t_freq[1])
      else if count = 19 then
      begin
         inc(t_freq[0]);
         inc(t_freq[1]);
      end else
         inc(t_freq[2]);
    end else { k<>0 }
      inc(t_freq[k+2]);
  end;
end; { count_t_freq }

{$R-} { ensure no spurious errors for next two routines... }
procedure TCompThreadSafety.putcode5(n: AnsiChar; x: Word);
begin
  while n >= bitcount do
  begin
    dec(n,ord(bitcount));
    inc(subbitbuf, x shr (WordBits - ord(bitcount)));
    x := x shl ord(bitcount);
    PutChar(subbitbuf);
{$IFDEF COMPARE_ON_COMPRESS}
    chkchar(ord(subbitbuf));
{$ENDIF}
    subbitbuf := #0;
    bitcount := AnsiChar(CHAR_BIT);
  end;
  inc(subbitbuf, x shr (WordBits - ord(bitcount)));
  dec(bitcount,ord(n));
end; { putcode }

procedure TCompThreadSafety.putbits(n: AnsiChar;x: Word);
begin
  x := x shl (WordBits-ord(n));
  putcode5(n,x);
end;


procedure TCompThreadSafety.write_position_len(n, nbit, i_special: Smallint);
var i,k: Smallint;
begin
  while (n > 0) and (position_len[pred(n)]=#0) do
    dec(n);
  putbits(AnsiChar(nbit),n);
  i := 0;
  while (i<n) do
  begin
    k := ord(position_len[i]); inc(i);
    if k <=6 then
       putbits(#3,k)
    else
       putbits(AnsiChar(k-3),Word(Longint(WordMax) shl 1));

    if i = i_special then
    begin
      while (i < 6) and (position_len[i] = #0) do
        inc(i);
      putbits(#2,i-3);
    end;
  end; { while }
end; { write_position_len }

procedure TCompThreadSafety.write_charLen;
var i, k, n, count: Smallint;
begin
  n := constNC;
  while (n > 0) and (charLen[pred(n)] = #0) do
    dec(n);
  putbits(AnsiChar(CBIT),n);
  i := 0;
  while i < n do
  begin
    k := ord(charLen[i]); inc(i);
    if k = 0 then
    begin
      count := 1;
      while (i < n) and (charLen[i] = #0) do
      begin
        inc(i);
        inc(count);
      end;
      if count <= 2 then
      begin
        for k := 0 to pred(count) do
          putcode5(position_len[0],position_code[0]);
      end else if count <= 18 then
      begin
        putcode5(position_len[1],position_code[1]);
        putbits(#4,count-3);
      end else if count = 19 then
      begin
        putcode5(position_len[0],position_code[0]);
        putcode5(position_len[1],position_code[1]);
        putbits(#4,15);
      end else
      begin
        putcode5(position_len[2],position_code[2]);
        putbits(AnsiChar(CBIT),count-20);
      end;
    end else { k <> 0 }
      putcode5(position_len[k+2],position_code[k+2]);
  end; { while i < n}
end; { write_charLen }

procedure TCompThreadSafety.encode_char(c: Smallint);
begin
  putcode5(charLen[c],char_code[c]);
end;

procedure TCompThreadSafety.encode_position(p: Word);
var c,q: Word;
begin
  c := 0;
  q := p;
  while q > 0 do
  begin
    q := q shr 1;
    inc(c);
  end;
  putcode5(position_len[c], position_code[c]);
  if c > 1 then
    putbits(AnsiChar(c-1),p);
end; { encode_p }

procedure TCompThreadSafety.send_block;
var flags: AnsiChar;
    i, k, root, pos, size: Word;
begin
  root := buildtree(constNC, char_freq,charLen,char_code);
  size := char_freq[root];
  putbits(#16,size);
  if root >= constNC then
  begin
    count_t_freq;
    root := buildtree(constNT,t_freq,position_len,position_code);
    if root >= constNT then
       write_position_len(constNT,TBIT,3)
    else begin
       putbits(AnsiChar(TBIT),0);
       putbits(AnsiChar(TBIT),root);
    end;
    write_charLen;
  end else { root < constNC }
  begin
    putbits(AnsiChar(TBIT),0);
    putbits(AnsiChar(TBIT),0);
    putbits(AnsiChar(CBIT),0);
    putbits(AnsiChar(CBIT),root);
  end; { if }

  root := buildtree(constNP,position_freq,position_len,position_code);
  if root >= constNP then
    write_position_len(constNP,PBIT,-1)
  else
  begin
    putbits(AnsiChar(PBIT),0);
    putbits(AnsiChar(PBIT),root);
  end;

  pos := 0; flags := #0;
  for i := 0 to pred(size) do
  begin
    if (i mod CHAR_BIT)= 0 then
    begin
      flags := buf[pos]; inc(pos);
    end else
      flags := AnsiChar(ord(flags) shl 1);

    if (ord(flags) and (1 shl pred(CHAR_BIT))) <> 0 then
    begin
      encode_char(Ord(buf[pos]) + (1 shl CHAR_BIT));
      inc(pos);
      k := ord(buf[pos]) shl CHAR_BIT; inc(pos);
      inc(k,ord(buf[pos])); inc(pos);
      encode_position(k);
    end else
    begin
      encode_char(ord(buf[pos])); inc(pos);
    end;
{ if unpackable then return }
  end; { for }

  for i := 0 to pred(constNC) do char_freq[i] := 0;
  for i := 0 to pred(constNP) do position_freq[i] := 0;
end; { send_block }

procedure TCompThreadSafety.OutputCodes(c, p: Word);
begin
  output_mask := output_mask shr 1;
  if output_mask = 0 then
  begin
    output_mask :=  1 shl pred(CHAR_BIT);
    if output_pos >= (BufferSize-3*CHAR_BIT) then
    begin
      send_block;
      output_pos := 0;
    end;
    cpos := output_pos;  inc(output_pos);
    buf[cpos] := #0;
  end;

  buf[output_pos] := AnsiChar(c); inc(output_pos);
  inc(char_freq[c]);

  if c >= (1 shl CHAR_BIT) then
  begin
    buf[cpos] := AnsiChar(ord(buf[cpos]) or output_mask);
    buf[output_pos] := AnsiChar(p shr CHAR_BIT); inc(output_pos);
    buf[output_pos] := AnsiChar(p); inc(output_pos);
    c := 0;
    while (p > 0) do
    begin
      p := p shr 1;
      inc(c);
    end;
    inc(position_freq[c]);
  end;
end;

procedure TCompThreadSafety.initSlide;
var i: node;
begin
  for i := DictionarySize to DictionarySize + CharMax do
  begin
    level[i] := #1;
    position[i] := NIL5;
  end;
  for i := DictionarySize to pred(DictionarySize * 2) do
     parent[i] := NIL5;
  avail := 1;
  for i := 1 to pred(DictionarySize-1) do
    next[i] := i+1;
  next[DictionarySize-1] := NIL5;
  for i := DictionarySize*2 to max_hash_val do
    next[i] := NIL5;
end;

function TCompThreadSafety.child(q: node; c: AnsiChar): node;
begin
  result := next[q + (ord(c) shl hash1) + hash2];
  parent[NIL5] := q;
  while parent[result] <> q do
    result := next[result];
end;

procedure TCompThreadSafety.makechild(q: node; c: AnsiChar; r: node);
var h,t: node;
begin
   h := q + (ord(c) shl hash1) + hash2;
   t := next[h];
   next[h] := r;
   next[r] := t;
   prev[t] := r;
   prev[r] := h;
   parent[r] := q;
   inc(childcount[q]);
end; { makechild }

procedure TCompThreadSafety.split(old: node);
var new, t: node;
begin
  new := avail;
  avail := next[new];
  childcount[new] := #0;

  t := prev[old];
  prev[new] := t;
  next[t] := new;

  t := next[old];
  next[new] := t;
  prev[t] := new;

  parent[new] := parent[old];
  level[new] := AnsiChar(MatchLength);
  position[new] := pos;

  makechild(new,text[MatchPosition+MatchLength],old);
  makechild(new,text[pos+MatchLength],pos);
end;

procedure TCompThreadSafety.insert_node;
var q,r,j,t: node;
    c: AnsiChar;
    ta,tb: Smallint;
begin
  q := 0; { silence compiler warning }
  if MatchLength >= 4 then
  begin
    dec(MatchLength);
    r := succ(MatchPosition) or DictionarySize;
    while True do
    begin
      q := parent[r];
      if q<>NIL5 then break;
      r := next[r];
    end;
    while ord(level[q]) >= MatchLength do
    begin
      r := q;
      q := parent[q];
    end;
    t := q;
    while position[t] < 0 do
    begin
      position[t] := pos;
      t := parent[t];
    end;
    if t < DictionarySize then
       position[t] := pos or SmallintMin;
  end else { MatchLength <= 4}
  begin
    q := ord(text[pos])+DictionarySize;
    c := text[pos+1];
    r := child(q,c);
    if r=NIL5 then
    begin
      makechild(q,c,pos);
      MatchLength := 1;
      exit;
    end;
    MatchLength := 2;
  end; { else }

  while True do
  begin
    if r >= DictionarySize then
    begin
      j := maxmatch;
      MatchPosition := r;
    end else
    begin
      j := ord(level[r]);
      MatchPosition := position[r] and SmallintMax;
    end;
    if MatchPosition >=pos then
       dec(MatchPosition,DictionarySize);
    ta := pos+MatchLength;
    tb := MatchPosition+MatchLength;
    while MatchLength < j do
    begin
      if text[ta]<>text[tb] then
      begin
        split(r);
        exit;
      end;
      inc(MatchLength);
      inc(ta);
      inc(tb);
    end; { while }
    if MatchLength=maxmatch then
      break;
    position[r] := pos;
    q := r;
    r := child(q,text[ta]);
    if r=NIL5 then
    begin
      makechild(q,text[ta],pos);
      exit;
    end;
    inc(MatchLength);
  end; { while true }

  prev[pos] := prev[r];
  next[prev[r]] := pos;

  next[pos] := next[r];
  prev[next[r]] := pos;

  parent[pos] := q;
  parent[r] := NIL5;
  next[r] := pos;
end; { insert_node}

procedure TCompThreadSafety.delete_node;
var q,r,s,t,u: node;
begin
  u := 0; { silence compiler warning }
  if parent[pos] = NIL5 then
     exit;

  next[prev[pos]] := next[pos];
  prev[next[pos]] := prev[pos];

  r := parent[pos];
  parent[pos] := NIL5;
  dec(childcount[r]);
  if (r >= DictionarySize) or (childcount[r] > #1) then
     exit;
  t := position[r] and SmallintMax;
  if t >= pos then
    dec(t,DictionarySize);

  s := t;
  q := parent[r];
  while True do
  begin
    u := position[q];
    if u >=0 then break;
    u := u and SmallintMax;
    if u >= pos then
       dec(u,DictionarySize);
    if u > s then
      s := u;
    position[q] := s or DictionarySize;
    q := parent[q];
  end; { while }

  if q < DictionarySize then
  begin
    if u >= pos then
      dec(u,DictionarySize);
    if u > s then
      s := u;
    position[q] := (s or DictionarySize) or SmallintMin;
  end;

  s := child(r,text[t+ord(level[r])]);
  t := prev[s];
  u := next[s];

  next[t] := u;
  prev[u] := t;

  t := prev[r];
  next[t] := s;
  prev[s] := t;

  t := next[r];
  prev[t] := s;
  next[s] := t;

  parent[s] := parent[r];
  parent[r] := NIL5;
  next[r] := avail;
  avail := r;
end; { delete_node }

procedure TCompThreadSafety.get_next_match;
var i: Smallint;
begin
  dec(remainder);
  inc(pos);
  if pos = DictionarySize*2 then
  begin
    { V8.0 - pointers no longer used, we have to MOVE bytes ourselves }
    for i := 0 to DictionarySize + MaxMatch - 1 do
      text[i] := text[DictionarySize + i];

    for i:= DictionarySize+maxmatch to pred(DictionarySize*2+maxmatch) do
    begin     { fill buffer with incoming data }
      text[i] := GetChar;
      if InputEOF then
         break;
    end;
    if inputEOF then
       inc(remainder,i-(DictionarySize+maxmatch)) { just what we got }
    else
       inc(remainder,DictionarySize);
    pos := DictionarySize;
  end;
  delete_node;
  insert_node;
end;

procedure TCompThreadSafety.CompressLZH5;
var  lastMatchLength : Smallint;
     lastMatchPosition: node;
     c: AnsiChar;
     i: node;
begin
{$IFDEF COMPARE_ON_COMPRESS}
     CompareFile := TFileStream.create('c:\tst.lzh',fmOpenRead);
     CompareFile.seek(32,soBeginning);
{$ENDIF}

 try
  StartHuff5(True); { this is Compress so LOTS more buffers required... }
  initSlide;

  for i := 0 to pred(constNC) do
    char_freq[i] := 0;
  for i := 0 to pred(constNP) do
    position_freq[i] := 0;
  buf[0] := #0;
  output_pos :=0;
  output_mask := 0;
  cpos := 0;

  pos := DictionarySize + maxmatch;
  remainder := 0;
  for i:= pos to pos+pred(DictionarySize) do { fill buffer with incoming data OR spaces... }
  begin
    c := GetChar;
    if InputEOF then
       text[i] := ' '
    else
    begin
       text[i] := c;
       inc(remainder);
    end;
  end;
  if remainder=0 then exit; { nothing to compress... }
  MatchPosition := 0;
  MatchLength := 0;
  insert_node;
  while remainder > 0 do
  begin
    lastMatchLength := MatchLength;
    lastMatchPosition := MatchPosition;

    get_next_match;
    if MatchLength > remainder then
      MatchLength := remainder;
    if (MatchLength > lastMatchLength) or (lastMatchLength < Threshold5) then
      OutputCodes(Word(text[pos-1]),0)
    else begin
      OutputCodes(lastMatchLength+CharMax + 1 - Threshold5,
                                  (pos - lastMatchPosition - 2) and pred(DictionarySize));
      for lastMatchLength := 1 to pred(lastMatchLength) do
         get_next_match;
      if MatchLength > remainder then
         MatchLength := remainder;
    end;     { else }
  end; { while }

{ if not unpackable then }
  send_block;
  putbits(AnsiChar(CHAR_BIT -1),0);
 finally
  EndHuff5; { free up stuff }
{$IFDEF COMPARE_ON_COMPRESS}
     CompareFile.free;
{$ENDIF}

 end;
end;

{$ENDIF} { LZH5_COMPRESS}

procedure TCompThreadSafety.CopyWithCRC;  { not hard, but slower than predecessor... }
var ch: AnsiChar;
begin
  while True do
  begin
    ch := GetChar;
    if InputEOF then break;
    Putchar(ch);
  end;
end;

function TCompress.ProcessStreams(outstream, instream: TStream; sourcesize,destsize: Int64;  { V8.0 upgraded from Longint }
         CompressionMethod: TCompressionMethod; var checksum: Longint; mode: TCProcessMode;
         encrypted: Boolean): Int64; { Encryption added V3.5 } { V8.0 upgraded from Longint }
var Keyval: Longint;
    progress: Longint;  { V8.0 keep it as Longint for backward compatibility with existing projects }
begin
  Global.GlobalFOnShowProgress := FOnShowProgress;  { save in global so Getchar can see }
  if assigned(FOnShowProgress) then
  begin
   Progress := 0;
   FOnShowProgress(Progress);
   if Progress = -1 then { they said QUIT so do nothing }
   begin
     result := 0;
     exit;
   end;
  end;
  with Global do
  begin
    source := inStream;
    dest := outStream;
    SetLength(inBuffer, ChunkSize);  { allocate the buffers }
    inBuffPos := 0;
    SetLength(outBuffer, ChunkSize); { allocate the buffers }
    outBuffPos := 0;
    InputEOF := False;
    lastch := #0;
    inRepeat := False;
    dupCount := 0;
    ReadSize := Sourcesize;
    lChunk := Chunksize;
    inChecksum:= 0;
    outChecksum := 0;
    BytesIn := 0; { for ShowProgress event }

    if mode=cmExpand then    { V3.5 }
      MaxNumberOfBytesToWrite := ConstMaxINT64  { permit any size expand! }
                                 { V8.0 updated from MaxLongInt}
    else   { on Compression, never write more than Sourcesize-overhead bytes }
     MaxNumberOfBytesToWrite := SourceSize-FMustCompressByAtLeast;

    EncryptingOut := (mode = cmCompress) and (Key<>0); { V3.5 }
    EncryptingIn := (mode = cmExpand) and (Key<>0) and
                    Encrypted;  { ignore Key unless flag is ALSO set in archive }
    Keyval := Key;  
    { V8.0 - pointers no longer used, we have to MOVE bytes ourselves }
    { V8.1 - but the Low/Hi bytes were incorrectly assigned in V8.0, so }
    EncryptCodes[0] := Byte(Keyval);
    EncryptCodes[1] := Byte(Keyval shr 8);
    EncryptCodes[2] := Byte(Keyval shr 16);
    EncryptCodes[3] := Byte(Keyval shr 24);
    EncryptCounter :=0;
    try
     if mode = cmCompress then
     begin
      BytesOut:=0;
      case CompressionMethod of
       coNone: CopyWithCRC; { straight copy, no expansion... }
       coRLE:
{$IFDEF RLE_COMPRESS}
        CompressRLE;
{$ELSE}
        raise EUnrecognizedCompressionMethod.Create(
                    'RLE compression disabled in this build');
{$ENDIF}

       coLZH:
{$IFDEF LZH1_COMPRESS}
        CompressLZH;
{$ELSE}
        raise EUnrecognizedCompressionMethod.Create(
                    'LZH1 compression disabled in this build');
{$ENDIF}

      coLZH5:
{$IFDEF LZH5_COMPRESS}
        CompressLZH5;
{$ELSE}
        raise EUnrecognizedCompressionMethod.Create(
                    'LZH5 compression disabled in this build');
{$ENDIF}

    end;
    checksum := InChecksum;
   end else { expand }
   begin
    BytesOut:=1;
    case CompressionMethod of
       coNone: CopyWithCRC; { straight copy, no expansion... }

       coRLE:
{$IFDEF RLE_EXPAND}
           ExpandRLE;
{$ELSE}
               raise EUnrecognizedCompressionMethod.Create(
                     'RLE expansion disabled in this build');
{$ENDIF}
       coLZH:
{$IFDEF LZH1_EXPAND}
           ExpandLZH(destsize);
{$ELSE}
               raise EUnrecognizedCompressionMethod.Create(
                     'LZH1 expansion disabled in this build');
{$ENDIF}

       coLZH5:
{$IFDEF LZH5_EXPAND}
           ExpandLZH5(destsize);
{$ELSE}
               raise EUnrecognizedCompressionMethod.Create(
                     'LZH5 expansion disabled in this build');
{$ENDIF}

      end;
      checksum := OutChecksum;
     end;
      if OutBuffPos <> 0 then { must flush }
{$IFDEF CLR}
        dest.WriteBuffer(OutBuffer, OutBuffPos);    // flush buffer
{$ELSE}
        dest.WriteBuffer(OutBuffer[0], OutBuffPos); // flush buffer
{$ENDIF}
    finally
      { free the buffers }
      SetLength(inBuffer, 0);
      SetLength(outBuffer, 0);
    end;
    Progress := 100;
    if assigned(FOnShowProgress) then FOnShowProgress(Progress);
    Result := BytesOut;
  end; { with Global }
end;

procedure Register;
begin
  RegisterComponents('Compress', [TCompress]);
end;

end.

